// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

#include <math.h>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include "Vector3.H"
#include "fixatorframe.h"

using namespace std;

int main(int argc, char **argv)
{
   int i, j, k, status, steps, stride;
   int holeCircleId[2];
   int ringHoleId[2][6];
   double rotationX, rotationY, rotationZ;
   double translationX, translationY, translationZ;
   string ringId, fileSuffix;
   RingParameters ringParameters[2];
   StrutParameters strutParameters[6];
   FixatorFrame frame;

   // get ring parameters from command line
   i = 1;
   ringId = argv[i++];
   for (j = 0; j < 2; j++) {
      ringParameters[j].innerRadius = atof(argv[i++]); 
      ringParameters[j].outerRadius =  atof(argv[i++]);
      ringParameters[j].thickness =  atof(argv[i++]);
      if (j == 0) 
         ringParameters[j].jointOffset = 10.2; 
      else
         ringParameters[j].jointOffset = -10.2;
      ringParameters[j].jointRadius = 6.0;
      ringParameters[j].supportRadius = 6.0;
      ringParameters[j].boneRadius = 3.0; 
      ringParameters[j].holeCircle[0].radius =  atof(argv[i++]);
      ringParameters[j].holeCircle[0].holeCount =  atoi(argv[i++]);
      ringParameters[j].holeCircle[1].radius = atof(argv[i++]);
      ringParameters[j].holeCircle[1].holeCount = atoi(argv[i++]);
      ringParameters[j].circleCount = 2;
      frame.ring[j].setParameters(j, ringParameters[j]);
   }

   // get strut parameters from command line
   for (j = 0; j < 6; j++) {
      strutParameters[j].minLength = atof(argv[i++]);
      strutParameters[j].maxLength = atof(argv[i++]);
      strutParameters[j].diameter = atof(argv[i++]);
      strutParameters[j].screwRadius = atof(argv[i++]);
      strutParameters[j].screwLength = atof(argv[i++]);
      strutParameters[j].nutRadius = atof(argv[i++]);
      strutParameters[j].nutLength = atof(argv[i++]);
      strutParameters[j].tubeRadius = atof(argv[i++]);
      strutParameters[j].tubeLength = atof(argv[i++]);
      strutParameters[j].jointRadius = atof(argv[i++]);
      strutParameters[j].jointOffset = atof(argv[i++]);
      strutParameters[j].clearance = 2;
      frame.strut[j].setParameters(j, strutParameters[j]);
   }

   // get strut lengths and approximate ring spacing from command line
   double strutLengths[6];
   double ringSpacing;
   for (j = 0; j < 6; j++) {
      strutLengths[j] = atof(argv[i++]);
   }
   ringSpacing = atof(argv[i++]);

   //get correction steps from command line
   steps = atoi(argv[i++]);

   // get file name suffix from command line
   fileSuffix = argv[i++];

#if 0
   // use 72 hole circle
   frame.strut[0].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 1, 1, 11);
   frame.strut[1].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 23, 1, 13);
   frame.strut[2].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 25, 1, 35);
   frame.strut[3].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 47, 1, 37);
   frame.strut[4].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 49, 1, 59);
   frame.strut[5].connectRings
      (&frame.ring[0], &frame.ring[1], 1, 71, 1, 61);
#endif

   // get hole circle parameters from command line
   for (j = 0; j < 2; j++)
      holeCircleId[j] = atoi(argv[i++]) - 1;  // 1-based -> 0-based
   for (j = 0; j < 6; j++) {
      ringHoleId[0][j] = atoi(argv[i++]) - 1;  // 1-based -> 0-based
      ringHoleId[1][j] = atoi(argv[i++]) - 1;
      frame.strut[j].connectRings(&frame.ring[0], &frame.ring[1], 
                                  holeCircleId[0], ringHoleId[0][j], 
                                  holeCircleId[1], ringHoleId[1][j]);
   }

   // configure bones (dummy)
   frame.ring[0].setLocalBoneCoords
      (Vector3(0.0, 0.0, -30.0), Vector3(10, 20, 90));
   frame.ring[1].setLocalBoneCoords
      (Vector3(0.0, 0.0, 50.0), Vector3(10, 20, -90));

   // set strut lengths
   status = frame.setStrutLengths(strutLengths, ringSpacing);
   if (status == -1) {  // not converging
      cout << "NOT CONVERGING" << endl;
      return -1;
   }

   // generate correction sequence -
   // ensure write directory has drwxrwxrwx permissions
   ofstream xml("corrections.xml");
   xml << fixed << setprecision(1);
   xml << "<correctionSchedule ringId=\"" << ringId << "\">" << endl;
   k = 0;
   stride = steps/5;  // divide step space into 6 for generating graphics
   for (i = 0; i < steps; i++) {
      if (i > 0)
         frame.correctDeformity(1/(double)(steps - i));

      // generate povray source and png graphics file
      if ((i == 0) || (i == steps - 1) || 
          ((i % stride) == 0)) {
         ostringstream pathName;
         pathName << "graphics/fixator" << k << fileSuffix << ".pov"; 
         ofstream povraySource(pathName.str().c_str());
         povraySource << "#include \"prefix.inc\"" << endl << endl;
         frame.emitPovraySource(povraySource, 1);
         povraySource.close();
         ostringstream command;
         command << "cd graphics; ";
         command << "povray -Ifixator" << k << fileSuffix << ".pov";
         command << " +A +D +W300 +H300";
         system(command.str().c_str());
         k++;
      }

      // output correction table
      xml << "  <correction id=\"" << i << "\"" << endl;
      xml << "   ";
      for (j = 0; j < 6; j++) {
         xml << " strut" << j + 1 <<"Length=\"";
         xml << frame.strut[j].length() << "\"";
         if (j == 2)
            xml << endl << "   ";
         else if (j == 5)
            xml << "/>" << endl;
      }
   }
   xml << "</correctionSchedule>" << endl;
   xml.close();

   return 0;
}



