// Copyright 2007 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// FixatorFrame - the assembled fixator frame

#include <math.h>
#include <values.h>
#include "fixatorframe.h"


FixatorFrame::FixatorFrame()
{
   // configure default rings
   RingParameters ringParameters;
   ringParameters.innerRadius = 75.0;
   ringParameters.outerRadius = 100.0;
   ringParameters.thickness = 7.5;
   ringParameters.jointOffset = 10.2;
   ringParameters.jointRadius = 6.0;
   ringParameters.supportRadius = 6.0;
   ringParameters.boneRadius = 3.0;
   ringParameters.holeCircle[0].radius = 82.0;
   ringParameters.holeCircle[0].holeCount = 42;
   ringParameters.holeCircle[1].radius = 93.0;
   ringParameters.holeCircle[1].holeCount = 48;
   ringParameters.circleCount = 2;
   ring[0].setParameters(0, ringParameters);
   ringParameters.jointOffset = -10.2;
   ring[1].setParameters(1, ringParameters);

   // configure default struts
   StrutParameters strutParameters;
   strutParameters.minLength = 116.0;
   strutParameters.maxLength = 178.0;
   strutParameters.diameter = 12.0;
   strutParameters.screwRadius = 4;
   strutParameters.screwLength = 100;
   strutParameters.nutRadius = 7; 
   strutParameters.nutLength = 10;
   strutParameters.tubeRadius = 7;
   strutParameters.tubeLength = 116;
   strutParameters.jointRadius = 6;
   strutParameters.jointOffset = 10.2;
   strutParameters.clearance = 2;
   for (int i = 0; i < 6; i++) {
      strut[i].setParameters(i, strutParameters);
   }

   // use 48 hole circle for default
   strut[0].connectRings(&ring[0], &ring[1], 1, 1, 1, 7);
   strut[1].connectRings(&ring[0], &ring[1], 1, 15, 1, 9);
   strut[2].connectRings(&ring[0], &ring[1], 1, 17, 1, 23);
   strut[3].connectRings(&ring[0], &ring[1], 1, 31, 1, 25);
   strut[4].connectRings(&ring[0], &ring[1], 1, 33, 1, 39);
   strut[5].connectRings(&ring[0], &ring[1], 1, 47, 1, 41);

   // configure bones (dummy)
   ring[0].setLocalBoneCoords(Vector3(0.0, 0.0, -30.0), Vector3(10, 20, 90));
   ring[1].setLocalBoneCoords(Vector3(0.0, 0.0, 50.0), Vector3(10, 20, -90));
}


// transformRingTo(): Absolute transformation.

void
FixatorFrame::transformRingTo
   (const int ringId, const Vector3 rotation, const Vector3 translation)
{
   ring[ringId].transformTo(rotation, translation);
}


// transformRingBy(): Relative transformation.

void
FixatorFrame::transformRingBy
   (const int ringId, const Vector3 rotation, const Vector3 translation)
{
   ring[ringId].transformBy(rotation, translation);
}


// interpolate on real line segment containing min and max -
// index identifies interpolation postition in range
double interpolate(int range, int index, double min, double max)
{
   return((index/double(range))*(max - min) + min);
}


int
FixatorFrame::setStrutLengths
   (const double strutLength[6], const double approximateZ)
{
   int index[6], target[6], range, iterations;
   double min[6], max[6], newMin, newMax;
   double precision, diff, squares, leastSquares;

   const int maxIterations = 100;
   const double tiny = 1.0e-1;

   // initial search space
   min[0] = min[1] = min[2] = -M_PI/4;  // x, y, z rotation
   max[0] = max[1] = max[2] = M_PI/4;
   min[3] = min[4] = -100.0;  // x, y translation
   max[3] = max[4] = 100.0;
   min[5] = approximateZ - 50;
   max[5] = approximateZ + 50;

   // search for least squares solution
   range = 5;
   iterations = 0;
   leastSquares = MAXDOUBLE;
   precision = 0.001;
   while (leastSquares > tiny) {
      iterations++;
      if (iterations > maxIterations) {
         cout << "not converging" << endl;
         return -1;
      }
      leastSquares = MAXDOUBLE;
      for (index[0] = 0; index[0] < range; index[0]++) {
         for (index[1] = 0; index[1] < range; index[1]++) {
            for (index[2] = 0; index[2] < range; index[2]++) {
               for (index[3] = 0; index[3] < range; index[3]++) {
                  for (index[4] = 0; index[4] < range; index[4]++) {
                     for (index[5] = 0; index[5] < range; index[5]++) {
                        transformRingTo
                          (1,
                           Vector3(
                              interpolate(range, index[0], min[0], max[0]), 
                              interpolate(range, index[1], min[1], max[1]), 
                              interpolate(range, index[2], min[2], max[2])),
                           Vector3(
                              interpolate(range, index[3], min[3], max[3]), 
                              interpolate(range, index[4], min[4], max[4]), 
                              interpolate(range, index[5], min[5], max[5])));
                        squares = 0.0;
                        for (int i = 0; i < 6; i++) {
                           diff = strutLength[i] - strut[i].length();
                           diff /= precision;
                           squares += diff*diff;
                        }
                        if (squares < leastSquares) {
                           leastSquares = squares;
                           for (int i = 0; i < 6; i++)
                              target[i] = index[i];
                        }
                     }
                  }
               }
            }
         }
      }
      for (int i = 0; i < 6; i++) {
         newMin = interpolate(range, target[i] - 1, min[i], max[i]);
         newMax = interpolate(range, target[i] + 1, min[i], max[i]);
         min[i] = newMin;
         max[i] = newMax;
      }

      cout << "leastSquares = " << leastSquares << endl;
      for (int i = 0; i < 3; i++) {
         cout << "val[i] = " 
              << interpolate(range, target[i], min[i], max[i])*180/M_PI << ", ";
      }
      for (int i = 3; i < 6; i++) {
         cout << "val[i] = " 
              << interpolate(range, target[i], min[i], max[i]) << ", ";
      }
      cout << "<br/>" << endl;
   }

   cout << "TRANSFORMATION = ";
   for (int i = 0; i < 3; i++) {
      cout << round(interpolate(
                 range, target[i], min[i], max[i])*(180.0/M_PI)/0.1)*0.1
           << " ";
   }
   for (int i = 3; i < 6; i++) {
      cout << round(interpolate(range, target[i], min[i], max[i])/0.1)*0.1
           << " ";
   }
   cout << "<br/>" << endl;

   return 0;
}


// correctDeformity(): This function provides progressive deformity correction.
// On each call the given fraction of the remaining deformity is corrected. So
// if the required total number of correction steps is 100, say, then 
// successive calls will use fraction arguments 1/100, 1/99, ... 1/2, 1.
void
FixatorFrame::correctDeformity(double fraction)
{
   Vector3 bone0 = ring[0].global.boneCoords[0] -
                       ring[0].global.boneCoords[1];
   Vector3 bone1 = ring[1].global.boneCoords[1] -
                       ring[1].global.boneCoords[0];

   // compute total rotation correction, normalising to +-PI
   Vector3 totalRotation = bone1.rotationsTo(bone0);
   for (int i = 0; i < 3; i++) {
      if (totalRotation[i] > 0.0) {
         while (totalRotation[i] > M_PI)
            totalRotation[i] -= 2*M_PI;
      } else {
         while (totalRotation[i] < -M_PI)
            totalRotation[i] += 2*M_PI;
      }
   }

   // apply fractional rotation and compute total translation correction
   Vector3 totalTranslation =
      ring[1].global.boneCoords[1] - ring[0].global.boneCoords[1];
   Vector3 targetCoords =
      ring[1].global.boneCoords[1] - totalTranslation*fraction;
   Vector3 rotation = totalRotation*fraction;
   transformRingBy(1, rotation, Vector3(0.0, 0.0, 0.0));
   Vector3 translation = targetCoords - ring[1].global.boneCoords[1];
   transformRingBy(1, Vector3(0.0, 0.0, 0.0), translation);
}


void
FixatorFrame::emitPovraySource(ofstream &povraySource, const bool showBones)
{
   for (int i = 0; i < 6; i++)
      strut[i].emitPovraySource(povraySource);
   for (int i = 0; i < 2; i++)
      ring[i].emitPovraySource(povraySource, showBones);
}
