// Copyright 2007 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// FixatorRing class

#include <math.h>
#include "primitive.h"
#include "artifact.h"
#include "fixatorring.h"
#include "fixatorstrut.h"


FixatorRing::FixatorRing()
{
   transformation = identity3D();
   local.centreCoords = Vector3(0.0, 0.0, 0.0);
   global.centreCoords = Vector3(0.0, 0.0, 0.0);
}


void
FixatorRing::setParameters(const int id, const RingParameters parameters)
{
   this->id = id;
   this->parameters = parameters;

   double theta;
   for (int circleId = 0; circleId < maxCircles; circleId++) {
      for (int holeId = 0; holeId < maxHoles; holeId++) {
         theta = 2*M_PI*holeId/parameters.holeCircle[circleId].holeCount;
         holeCoords[circleId][holeId][0] = 
            parameters.holeCircle[circleId].radius*cos(theta);
         holeCoords[circleId][holeId][1] = 
            parameters.holeCircle[circleId].radius*sin(theta);
         holeCoords[circleId][holeId][2] = 0.0;
      }
   }
}


void
FixatorRing::connectStrut(FixatorStrut *strut, 
                          const int circleId, const int holeId)
{
   int strutId = strut->id;
   this->strut[strutId] = strut;

   local.ringCoords[strutId] = holeCoords[circleId][holeId];
   local.jointCoords[strutId] = holeCoords[circleId][holeId];
   local.jointCoords[strutId][2] += parameters.jointOffset;
   global.ringCoords[strutId] = transformation*local.ringCoords[strutId];
   global.jointCoords[strutId] = transformation*local.jointCoords[strutId];
}


void
FixatorRing::setLocalBoneCoords(const Vector3 point0, const Vector3 point1)
{
   local.boneCoords[0] = point0;
   local.boneCoords[1] = point1;
   global.boneCoords[0] = transformation*local.boneCoords[0];
   global.boneCoords[1] = transformation*local.boneCoords[1];
}


// TransformTo(): Absolute transformation. Note that the coordinate 
// transformation expression must be in the order 'transformation*coordinates'.
// This is not commutative.

void
FixatorRing::transformTo(const Vector3 rotation, const Vector3 translation)
{
   transformation = Matrix4(identity3D(), rotation, translation);

   // transform coordinates
   global.centreCoords = transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = transformation*local.ringCoords[i];
      global.jointCoords[i] = transformation*local.jointCoords[i];
   }
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = transformation*local.boneCoords[i];
   }
}


// TransformBy(): Relative transformation. Note that the coordinate 
// transformation expression must be in the order 'transformation*coordinates'.
// This is not commutative.

void
FixatorRing::transformBy(const Vector3 rotation, const Vector3 translation)
{
   transformation = Matrix4(transformation, rotation, translation);

   // transform coordinates
   global.centreCoords = transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = transformation*local.ringCoords[i];
      global.jointCoords[i] = transformation*local.jointCoords[i];
   }
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = transformation*local.boneCoords[i];
   }
}


// emitPovraySource(): The model is constructed in local coordinate space,
// then transformed to global when emitPovraySource() for the ring is called.

void
FixatorRing::emitPovraySource(ofstream &povraySource, const bool showBone)
{
   Vector3 supportCoords[6];
   Artifact ring("ring");
   OpenCylinder body;
   Cylinder support[6];
   Sphere joint[6];
   Cylinder bone;

   for (int i = 0; i < 6; i++) {
      supportCoords[i][0] = local.jointCoords[i][0];
      supportCoords[i][1] = local.jointCoords[i][1];
      if (id == 0)
         supportCoords[i][2] = local.jointCoords[i][2] - 3;  // TEMP
      else
         supportCoords[i][2] = local.jointCoords[i][2] + 3;  // TEMP
      // if these are removed, povray fails with degenerate cylinder issue ??
   }

   if (id == 0) 
      body = OpenCylinder(Vector3(0.0, 0.0, 0.0), 
                          Vector3(0.0, 0.0, -parameters.thickness),
                          parameters.outerRadius, parameters.innerRadius);
   else
      body = OpenCylinder(Vector3(0.0, 0.0, 0.0),
                          Vector3(0.0, 0.0, parameters.thickness),
                          parameters.outerRadius, parameters.innerRadius);
   ring.add(&body);
   for (int i = 0; i < 6; i++) {
      support[i] = Cylinder(local.ringCoords[i], 
                            supportCoords[i],
                            parameters.supportRadius);
      ring.add(&support[i]);
      joint[i] = Sphere(local.jointCoords[i], parameters.jointRadius);
      ring.add(&joint[i]);
   }
   if (showBone) {
      bone = Cylinder(local.boneCoords[0], 
                      local.boneCoords[1],
                      parameters.boneRadius);
      ring.add(&bone);
   }

   ring.emitPovraySource(povraySource, transformation);
}
