
// C++ Vector and Matrix Algebra Routines from Graphics Gems IV
// Author: Jean-Francois DOUE
// Version 3.1 --- October 1993
// Minor modifications by John Storrs

#include <math.h>
#include <iostream>
#include <fstream>
#include <ctype.h>

#include "Vector3.H"
#include "Vector4.H"
#include "Matrix4.H"

using namespace std;

// Matrix4 Class Member Functions

// CONSTRUCTORS

Matrix4::Matrix4() {}

Matrix4::Matrix4
  (const Vector4& v0, const Vector4& v1, 
   const Vector4& v2, const Vector4& v3)
{ v[0] = v0; v[1] = v1; v[2] = v2; v[3] = v3; }

Matrix4::Matrix4
  (const double d)
{ v[0] = v[1] = v[2] = v[3] = Vector4(d); }

Matrix4::Matrix4
  (const Matrix4& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3]; }

//JS addition
Matrix4::Matrix4
  (const Matrix4& matrix, 
   const Vector3& rotation, const Vector3& translation)
// concatenate matrix and transformations, in order RX, RY, RZ, T
// very crude implementation using rotation3D - to be improved
{
   Matrix4 newMatrix = matrix;
   Matrix4 translationXYZ = translation3D(translation);
   Vector3 xAxis = Vector3(1, 0, 0);
   Vector3 yAxis = Vector3(0, 1, 0);
   Vector3 zAxis = Vector3(0, 0, 1);
   Matrix4 rotationX = rotation3D(xAxis, rotation[VX]);
   Matrix4 rotationY = rotation3D(yAxis, rotation[VY]);
   Matrix4 rotationZ = rotation3D(zAxis, rotation[VZ]);
   newMatrix = rotationX * newMatrix;
   newMatrix = rotationY * newMatrix;
   newMatrix = rotationZ * newMatrix;
   newMatrix = translationXYZ * newMatrix;

   *this = newMatrix;
}

// ASSIGNMENT OPERATORS

Matrix4& Matrix4::operator = ( const Matrix4& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3];
return *this; }

Matrix4& Matrix4::operator += ( const Matrix4& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; v[3] += m.v[3];
return *this; }

Matrix4& Matrix4::operator -= ( const Matrix4& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; v[3] -= m.v[3];
return *this; }

Matrix4& Matrix4::operator *= ( const double d )
{ v[0] *= d; v[1] *= d; v[2] *= d; v[3] *= d; return *this; }

Matrix4& Matrix4::operator /= ( const double d )
{ v[0] /= d; v[1] /= d; v[2] /= d; v[3] /= d; return *this; }

Vector4& Matrix4::operator [] ( int i) {
    if (i < VX || i > VW)
	V_ERROR("Matrix4 [] operator: illegal access; index = " << i << '\n')
    return v[i];
}

// SPECIAL FUNCTIONS;

Matrix4 Matrix4::transpose() {
    return Matrix4
      (Vector4(v[0][0], v[1][0], v[2][0], v[3][0]),
       Vector4(v[0][1], v[1][1], v[2][1], v[3][1]),
       Vector4(v[0][2], v[1][2], v[2][2], v[3][2]),
       Vector4(v[0][3], v[1][3], v[2][3], v[3][3]));
}

Matrix4 Matrix4::inverse()  // Gauss-Jordan elimination with partial pivoting
{
    Matrix4 a(*this),  // As a evolves from original mat into identity
	 b(identity3D());  // b evolves from identity into inverse(a)
    int i, j, i1;

    // Loop over cols of a from left to right, eliminating above and below diag
    for (j=0; j<4; j++) {   // Find largest pivot in column j among rows j..3
    i1 = j;  // Row with largest pivot candidate
    for (i=j+1; i<4; i++)
	if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
	   i1 = i;

    // Swap rows i1 and j in a and b to put pivot on diagonal
    swap(a.v[i1], a.v[j]);
    swap(b.v[i1], b.v[j]);

    // Scale row j to have a unit diagonal
    if (a.v[j].n[j]==0.)
	V_ERROR("Matrix4::inverse: singular matrix; can't invert\n");
    b.v[j] /= a.v[j].n[j];
    a.v[j] /= a.v[j].n[j];

    // Eliminate off-diagonal elems in col j of a, doing identical ops to b
    for (i=0; i<4; i++)
	if (i!=j) {
	b.v[i] -= a.v[i].n[j]*b.v[j];
	a.v[i] -= a.v[i].n[j]*a.v[j];
	}
    }
    return b;
}

Matrix4& Matrix4::apply(V_FCT_PTR fct)
{ v[VX].apply(fct); v[VY].apply(fct); v[VZ].apply(fct); v[VW].apply(fct);
return *this; }


// FRIENDS

Matrix4 operator - (const Matrix4& a)
{ return Matrix4(-a.v[0], -a.v[1], -a.v[2], -a.v[3]); }

Matrix4 operator + (const Matrix4& a, const Matrix4& b)
{ return Matrix4(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2],
  a.v[3] + b.v[3]);
}

Matrix4 operator - (const Matrix4& a, const Matrix4& b)
{ return Matrix4(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2], a.v[3] - b.v[3]); }

Matrix4 operator * (Matrix4& a, Matrix4& b) {
    #define ROWCOL(i, j) a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + \
    a.v[i].n[2]*b.v[2][j] + a.v[i].n[3]*b.v[3][j]
    return Matrix4(
    Vector4(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2), ROWCOL(0,3)),
    Vector4(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2), ROWCOL(1,3)),
    Vector4(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2), ROWCOL(2,3)),
    Vector4(ROWCOL(3,0), ROWCOL(3,1), ROWCOL(3,2), ROWCOL(3,3))
    );
}

Matrix4 operator * (const Matrix4& a, const double d)
{ return Matrix4(a.v[0] * d, a.v[1] * d, a.v[2] * d, a.v[3] * d); }

Matrix4 operator * (const double d, const Matrix4& a)
{ return a*d; }

Matrix4 operator / (const Matrix4& a, const double d)
{ return Matrix4(a.v[0] / d, a.v[1] / d, a.v[2] / d, a.v[3] / d); }

int operator == (const Matrix4& a, const Matrix4& b)
{ return ((a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]) &&
  (a.v[3] == b.v[3])); }

int operator != (const Matrix4& a, const Matrix4& b)
{ return !(a == b); }

ostream& operator << (ostream& s, Matrix4& m)
{ return s << m.v[VX] << '\n' << m.v[VY] << '\n' << m.v[VZ] << '\n' << m.v[VW]; }

istream& operator >> (istream& s, Matrix4& m)
{
    Matrix4    m_tmp;

    s >> m_tmp[VX] >> m_tmp[VY] >> m_tmp[VZ] >> m_tmp[VW];
    if (s)
	m = m_tmp;
    return s;
}

void swap(Matrix4& a, Matrix4& b)
{ Matrix4 tmp(a); a = b; b = tmp; }


// FUNCTIONS

Matrix4 identity3D()
{   return Matrix4
      (Vector4(1.0, 0.0, 0.0, 0.0),
       Vector4(0.0, 1.0, 0.0, 0.0),
       Vector4(0.0, 0.0, 1.0, 0.0),
       Vector4(0.0, 0.0, 0.0, 1.0)); }

Matrix4 translation3D(const Vector3& v)

{   return Matrix4
      (Vector4(1.0, 0.0, 0.0, v[VX]),
       Vector4(0.0, 1.0, 0.0, v[VY]),
       Vector4(0.0, 0.0, 1.0, v[VZ]),
       Vector4(0.0, 0.0, 0.0, 1.0)); }

Matrix4 rotation3D(const Vector3& Axis, const double angle) {
    double  
       c = cos(angle),
       s = sin(angle),
       t = 1.0 - c;

    Vector3 axisCopy = Axis;  // fix const problem (normalise() changes 'this')

    axisCopy.normalize();
    return Matrix4
      (Vector4
         (t * axisCopy[VX] * axisCopy[VX] + c,
          t * axisCopy[VX] * axisCopy[VY] - s * axisCopy[VZ],
          t * axisCopy[VX] * axisCopy[VZ] + s * axisCopy[VY], 0.0),
       Vector4
         (t * axisCopy[VX] * axisCopy[VY] + s * axisCopy[VZ],
	  t * axisCopy[VY] * axisCopy[VY] + c,
	  t * axisCopy[VY] * axisCopy[VZ] - s * axisCopy[VX], 0.0),
       Vector4
         (t * axisCopy[VX] * axisCopy[VZ] - s * axisCopy[VY],
	  t * axisCopy[VY] * axisCopy[VZ] + s * axisCopy[VX],
	  t * axisCopy[VZ] * axisCopy[VZ] + c, 0.0),
       Vector4(0.0, 0.0, 0.0, 1.0));
}

Matrix4 scaling3D(const Vector3& scaleVector)
{   return Matrix4
      (Vector4(scaleVector[VX], 0.0, 0.0, 0.0),
       Vector4(0.0, scaleVector[VY], 0.0, 0.0),
       Vector4(0.0, 0.0, scaleVector[VZ], 0.0),
       Vector4(0.0, 0.0, 0.0, 1.0)); }

Matrix4 perspective3D(const double d)
{   return Matrix4
      (Vector4(1.0, 0.0, 0.0, 0.0),
       Vector4(0.0, 1.0, 0.0, 0.0),
       Vector4(0.0, 0.0, 1.0, 0.0),
       Vector4(0.0, 0.0, 1.0/d, 0.0)); }

