// name:     computecorrections.cpp
// function: Computes a hexapod fixator correction schedule.
// author:   John Storrs, Laboratory for Micro Enterprise.
// license:  Released under the terms of the GNU General Public License.
// date:     $Date: 2008-05-24 17:08:31 +0100 (Sat, 24 May 2008) $

#include <math.h>
#include <float.h>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include "Vector2.H"
#include "Vector3.H"
#include "Matrix4.H"
#include "fixatorframe.h"

using namespace std;

void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight);


int main(int argc, char **argv)
{
   int i, j, k, status, steps, stride;
   int ringHoleId[2][6];
   double ringNearest[2];
   double rotationX, rotationY, rotationZ;
   double translationX, translationY, translationZ;
   string frameId, fileSuffix;
   RingParameters ringParameters[2];
   Matrix4 transformation;
   StrutParameters strutParameters[6];
   FixatorFrame frame;
   XrayParameters xrayParameters;
   DeformityParameters deformityParameters;
   GraphicsParameters graphicsParameters;

   // get frame id from command line
   i = 1;
   frameId = argv[i++];

   // get ring parameters from command line
   for (j = 0; j < 2; j++) {
      ringParameters[j].innerRadius = atof(argv[i++]); 
      ringParameters[j].outerRadius =  atof(argv[i++]);
      ringParameters[j].thickness =  atof(argv[i++]);
      if (j == 0) 
         ringParameters[j].jointOffset = 10.2; 
      else
         ringParameters[j].jointOffset = -10.2;
      ringParameters[j].jointRadius = 6.0;
      ringParameters[j].supportRadius = 6.0;
      ringParameters[j].boneRadius = 2.0; 
      ringParameters[j].holeCircleRadius =  atof(argv[i++]);
      ringParameters[j].holeCount =  atoi(argv[i++]);
      frame.ring[j].setParameters(j, ringParameters[j]);
   }

   // get strut parameters from command line
   for (j = 0; j < 6; j++) {
      strutParameters[j].minLength = atof(argv[i++]);
      strutParameters[j].maxLength = atof(argv[i++]);
      strutParameters[j].diameter = atof(argv[i++]);
      strutParameters[j].screwRadius = atof(argv[i++]);
      strutParameters[j].screwLength = atof(argv[i++]);
      strutParameters[j].nutRadius = atof(argv[i++]);
      strutParameters[j].nutLength = atof(argv[i++]);
      strutParameters[j].tubeRadius = atof(argv[i++]);
      strutParameters[j].tubeLength = atof(argv[i++]);
      strutParameters[j].jointRadius = atof(argv[i++]);
      strutParameters[j].jointOffset = atof(argv[i++]);
      strutParameters[j].clearance = 2;
      frame.strut[j].setParameters(j, strutParameters[j]);
   }

   // get required strut length parameters from command line
   for (j = 0; j < 4; j++)
      for (k = 0; k < 4; k++)
         transformation[j][k] = atof(argv[i++]);
   frame.ring[0].transformTo(transformation);
   for (j = 0; j < 4; j++)
      for (k = 0; k < 4; k++)
         transformation[j][k] = atof(argv[i++]);
   frame.ring[1].transformTo(transformation);

   // get xray parameters from command line
   xrayParameters.xrayDistanceFrontal = atof(argv[i++]);
   xrayParameters.ring0NearestFrontal = atof(argv[i++]);
   xrayParameters.ring1NearestFrontal = atof(argv[i++]);
   xrayParameters.xrayDistanceLateral = atof(argv[i++]);
   xrayParameters.ring0NearestLateral = atof(argv[i++]);
   xrayParameters.ring1NearestLateral = atof(argv[i++]);
   frame.setXrayParameters(xrayParameters);

   // get deformity parameters from command line
   deformityParameters.x1Ring1Frontal = atof(argv[i++]);
   deformityParameters.z1Ring1Frontal = atof(argv[i++]);
   deformityParameters.x2Ring1Frontal = atof(argv[i++]);
   deformityParameters.z2Ring1Frontal = atof(argv[i++]);
   deformityParameters.x1Ring2Frontal = atof(argv[i++]);
   deformityParameters.z1Ring2Frontal = atof(argv[i++]);
   deformityParameters.x2Ring2Frontal = atof(argv[i++]);
   deformityParameters.z2Ring2Frontal = atof(argv[i++]);
   deformityParameters.x1Ring1Lateral = atof(argv[i++]);
   deformityParameters.z1Ring1Lateral = atof(argv[i++]);
   deformityParameters.x2Ring1Lateral = atof(argv[i++]);
   deformityParameters.z2Ring1Lateral = atof(argv[i++]);
   deformityParameters.x1Ring2Lateral = atof(argv[i++]);
   deformityParameters.z1Ring2Lateral = atof(argv[i++]);
   deformityParameters.x2Ring2Lateral = atof(argv[i++]);
   deformityParameters.z2Ring2Lateral = atof(argv[i++]);
   deformityParameters.zRotation = atof(argv[i++]);
   frame.setDeformityParameters(deformityParameters);

   // get required frame parameters from command line
   for (j = 0; j < 6; j++) {
      ringHoleId[0][j] = atoi(argv[i++]) - 1;  // 1-based -> 0-based
      ringHoleId[1][j] = atoi(argv[i++]) - 1;
      frame.strut[j].connectRings(&frame.ring[0], &frame.ring[1], 
                                  ringHoleId[0][j], ringHoleId[1][j]);
   }

   // get correction steps from command line
   steps = atoi(argv[i++]);

   // get file name suffix from command line
   fileSuffix = argv[i++];

   // compute bone reference coordinates
   frame.computeBoneCoords();

   // set display parameters - TEMPORARY
   frame.showOrigin = 0;
   frame.showTarget = 0;
   frame.showNearest = 0;
   frame.showGrid = 0;
   frame.showBones = 1;

   // set graphics parameters
   graphicsParameters.showOrigin = 0;
   graphicsParameters.showTarget = 0;
   graphicsParameters.showNearest = 0;
   graphicsParameters.showGrid = 0;
   graphicsParameters.showBones = 1;
   graphicsParameters.imageWidth = 280;  // mm
   graphicsParameters.imageHeight = 380;  // mm
   graphicsParameters.pixelsPerInch = 150;
   frame.setGraphicsParameters(graphicsParameters);
   double widthPixels =
      graphicsParameters.imageWidth*graphicsParameters.pixelsPerInch/25.4;
   double heightPixels =
      graphicsParameters.imageHeight*graphicsParameters.pixelsPerInch/25.4;

   // transform frame to frontal view
   ringNearest[0] = xrayParameters.ring0NearestFrontal;
   ringNearest[1] = xrayParameters.ring1NearestFrontal;
   frame.transformView(ringNearest);

   // generate correction sequence -
   // ensure write directory has drwxrwxrwx permissions
   ofstream xml("corrections.xml");
   xml << fixed << setprecision(1);
   xml << "<correctionSchedule frameId=\"" << frameId << "\">" << endl;
   k = 0;
   stride = steps/5;  // divide step space into 6 for generating graphics
   for (i = 0; i < steps; i++) {
      if (i > 0)
         frame.correctDeformity(1/(double)(steps - i));

      // generate povray source and png graphics file
      if ((i == 0) || (i == steps - 1) || 
          ((i % stride) == 0)) {

         // generate frontal view graphics
         ostringstream pathName;
         pathName << "graphics/fixator" << k << fileSuffix << ".pov"; 
         ofstream povraySource(pathName.str().c_str());
         emitPovraySource(povraySource, xrayParameters.xrayDistanceFrontal,
                          graphicsParameters.imageWidth,
                          graphicsParameters.imageHeight);
         frame.emitPovraySource(povraySource);
         povraySource.close();
         ostringstream command;
         command << "cd graphics; ";
         command << "povray -Ifixator" << k << fileSuffix << ".pov";
         command << " +A +D +Q0 +W280 +H380";
         system(command.str().c_str());
         k++;
      }

      // output correction table
      xml << "  <correction id=\"" << i + 1 << "\"" << endl;
      xml << "   ";
      for (j = 0; j < 6; j++) {
         xml << " strut" << j + 1 <<"Length=\"";
         xml << frame.strut[j].length() << "\"";
         if (j == 2)
            xml << endl << "   ";
         else if (j == 5)
            xml << "/>" << endl;
      }
   }
   xml << "</correctionSchedule>" << endl;
   xml.close();

   return 0;
}


void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight)
{
   povraySource << "global_settings { assumed_gamma 2.2 }" << endl;
   povraySource << "#include \"colors.inc\"" << endl;
   povraySource << "#include \"metals.inc\"" << endl;
   povraySource << "camera {" << endl;
   povraySource << "  perspective" << endl;
   povraySource << "  location <0.0, -" << xrayDistance << ", 0.0>" << endl;
   povraySource << "  direction <0.0, " << xrayDistance << ", 0.0>" << endl;
   povraySource << "  up <0, 0, " << imageHeight << ">" << endl;
   povraySource << "  right <" 
                << imageWidth << ", 0, 0>" << endl;  // right hand rule
   povraySource << "  sky <0, 0, 1>" << endl;
   povraySource << "  angle " 
                << 2*atan2(imageWidth/2, xrayDistance)*180/M_PI << endl;
   povraySource << "  look_at <0.0, 0.0, 0.0>" << endl;
   povraySource << "}" << endl;
   povraySource << "background { color red 1 green 1 blue 1 }" << endl;
   povraySource << "light_source { <-2000, -2000, 0> color Gray60}" << endl;
   povraySource << "#declare FixatorStrut =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.5 green 0.5 blue 0.5" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorGrid =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.3 green 0.3 blue 0.3" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorAxes =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorRing =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.8 green 0.8 blue 0.8" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorNearest =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 0 blue 1" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorTarget =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorOrigin=" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorBone =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
}
