
// C++ Vector and Matrix Algebra Routines from Graphics Gems IV
// Author: Jean-Francois DOUE
// Version 3.1 --- October 1993
// Minor modifications by John Storrs

#include <math.h>
#include <iostream>
#include <fstream>
#include <ctype.h>

#include "Vector2.H"
#include "Vector3.H"
#include "Matrix3.H"

using namespace std;

// Vector2 Class Member Functions

// CONSTRUCTORS

Vector2::Vector2() {}

Vector2::Vector2(const double x, const double y)
{ n[VX] = x; n[VY] = y; }

Vector2::Vector2(const double d)
{ n[VX] = n[VY] = d; }

Vector2::Vector2(const Vector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; }

Vector2::Vector2(const Vector3& v) // it is up to caller to avoid divide-by-zero
{ n[VX] = v.n[VX]/v.n[VZ]; n[VY] = v.n[VY]/v.n[VZ]; };

Vector2::Vector2(const Vector3& v, int dropAxis) {
    switch (dropAxis) {
	case VX: n[VX] = v.n[VY]; n[VY] = v.n[VZ]; break;
	case VY: n[VX] = v.n[VX]; n[VY] = v.n[VZ]; break;
	default: n[VX] = v.n[VX]; n[VY] = v.n[VY]; break;
    }
}


// ASSIGNMENT OPERATORS

Vector2& Vector2::operator = (const Vector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; return *this; }

Vector2& Vector2::operator += ( const Vector2& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; return *this; }

Vector2& Vector2::operator -= ( const Vector2& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; return *this; }

Vector2& Vector2::operator *= ( const double d )
{ n[VX] *= d; n[VY] *= d; return *this; }

Vector2& Vector2::operator /= ( const double d )
{ double d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; return *this; }

double& Vector2::operator [] ( int i) {
    if (i < VX || i > VY)
	V_ERROR("Vector2 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

double Vector2::length()
{ return sqrt(length2()); }

double Vector2::length2()
{ return n[VX]*n[VX] + n[VY]*n[VY]; }

Vector2& Vector2::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }

Vector2& Vector2::apply(V_FCT_PTR fct)
{ n[VX] = (*fct)(n[VX]); n[VY] = (*fct)(n[VY]); return *this; }


// FRIENDS

Vector2 operator - (const Vector2& a)
{ return Vector2(-a.n[VX],-a.n[VY]); }

Vector2 operator + (const Vector2& a, const Vector2& b)
{ return Vector2(a.n[VX]+ b.n[VX], a.n[VY] + b.n[VY]); }

Vector2 operator - (const Vector2& a, const Vector2& b)
{ return Vector2(a.n[VX]-b.n[VX], a.n[VY]-b.n[VY]); }

Vector2 operator * (const Vector2& a, const double d)
{ return Vector2(d*a.n[VX], d*a.n[VY]); }

Vector2 operator * (const double d, const Vector2& a)
{ return a*d; }

Vector2 operator * (const Matrix3& a, const Vector2& v) {
    Vector3 av;

    av.n[VX] = a.v[0].n[VX]*v.n[VX] + a.v[0].n[VY]*v.n[VY] + a.v[0].n[VZ];
    av.n[VY] = a.v[1].n[VX]*v.n[VX] + a.v[1].n[VY]*v.n[VY] + a.v[1].n[VZ];
    av.n[VZ] = a.v[2].n[VX]*v.n[VX] + a.v[2].n[VY]*v.n[VY] + a.v[2].n[VZ];
    return av;
}

Vector2 operator * (const Vector2& v, Matrix3& a)
{ return a.transpose() * v; }

double operator * (const Vector2& a, const Vector2& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY]); }

Vector2 operator / (const Vector2& a, const double d)
{ double d_inv = 1./d; return Vector2(a.n[VX]*d_inv, a.n[VY]*d_inv); }

Vector3 operator ^ (const Vector2& a, const Vector2& b)
{ return Vector3(0.0, 0.0, a.n[VX] * b.n[VY] - b.n[VX] * a.n[VY]); }

int operator == (const Vector2& a, const Vector2& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]); }

int operator != (const Vector2& a, const Vector2& b)
{ return !(a == b); }

ostream& operator << (ostream& s, Vector2& v)
{ return s << "| " << v.n[VX] << ' ' << v.n[VY] << " |"; }

istream& operator >> (istream& s, Vector2& v) {
    Vector2	v_tmp;
    char	c = ' ';

    while (isspace(c))
	s >> c;
    // The vectors can be formatted either as x y or | x y |
    if (c == '|') {
	s >> v_tmp[VX] >> v_tmp[VY];
	while (s >> c && isspace(c)) ;
//	if (c != '|')
//	    s.set(_bad);     JS NOTE: _bad causes error when fstream.h included
	}
    else {
	s.putback(c);
	s >> v_tmp[VX] >> v_tmp[VY];
	}
    if (s)
	v = v_tmp;
    return s;
}

void swap(Vector2& a, Vector2& b)
{ Vector2 tmp(a); a = b; b = tmp; }

Vector2 min(const Vector2& a, const Vector2& b)
{ return Vector2(MIN(a.n[VX], b.n[VX]), MIN(a.n[VY], b.n[VY])); }

Vector2 max(const Vector2& a, const Vector2& b)
{ return Vector2(MAX(a.n[VX], b.n[VX]), MAX(a.n[VY], b.n[VY])); }

Vector2 prod(const Vector2& a, const Vector2& b)
{ return Vector2(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY]); }

