
// C++ Vector and Matrix Algebra Routines from Graphics Gems IV
// Author: Jean-Francois DOUE
// Version 3.1 --- October 1993
// Minor modifications by John Storrs

#include <math.h>
#include <iostream>
#include <fstream>
#include <ctype.h>

#include "Vector2.H"
#include "Vector3.H"
#include "Vector4.H"
#include "Matrix4.H"

using namespace std;

// Vector3 Class Member Functions			       *

// CONSTRUCTORS

Vector3::Vector3() {}

Vector3::Vector3(const double x, const double y, const double z)
{ n[VX] = x; n[VY] = y; n[VZ] = z; }

Vector3::Vector3(const double d)
{ n[VX] = n[VY] = n[VZ] = d; }

Vector3::Vector3(const Vector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; }

Vector3::Vector3(const Vector2& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = 1.0; }

Vector3::Vector3(const Vector2& v, double d)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = d; }

Vector3::Vector3(const Vector4& v) // it is up to caller to avoid divide-by-zero
{ n[VX] = v.n[VX] / v.n[VW]; n[VY] = v.n[VY] / v.n[VW];
  n[VZ] = v.n[VZ] / v.n[VW]; }

Vector3::Vector3(const Vector4& v, int dropAxis) {
    switch (dropAxis) {
	case VX: n[VX] = v.n[VY]; n[VY] = v.n[VZ]; n[VZ] = v.n[VW]; break;
	case VY: n[VX] = v.n[VX]; n[VY] = v.n[VZ]; n[VZ] = v.n[VW]; break;
	case VZ: n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VW]; break;
	default: n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; break;
    }
}


// ASSIGNMENT OPERATORS

Vector3& Vector3::operator = (const Vector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; return *this; }

Vector3& Vector3::operator += ( const Vector3& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; n[VZ] += v.n[VZ]; return *this; }

Vector3& Vector3::operator -= ( const Vector3& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; n[VZ] -= v.n[VZ]; return *this; }

Vector3& Vector3::operator *= ( const double d )
{ n[VX] *= d; n[VY] *= d; n[VZ] *= d; return *this; }

Vector3& Vector3::operator /= ( const double d )
{ double d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; n[VZ] *= d_inv;
  return *this; }

double& Vector3::operator [] ( int i) {
    if (i < VX || i > VZ)
	V_ERROR("Vector3 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}

double Vector3::operator [] ( int i) const {
    if (i < VX || i > VZ)
	V_ERROR("Vector3 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

double Vector3::length()
{  return sqrt(length2()); }

double Vector3::length2()
{  return n[VX]*n[VX] + n[VY]*n[VY] + n[VZ]*n[VZ]; }

Vector3& Vector3::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }

Vector3& Vector3::apply(V_FCT_PTR fct)
{ n[VX] = (*fct)(n[VX]); n[VY] = (*fct)(n[VY]); n[VZ] = (*fct)(n[VZ]);
return *this; }


// JS ADDITIONS

Vector3
Vector3::translate(const Vector3 translations)
// point *this is translated
{
   return Vector3
      (n[0] + translations.n[0], 
       n[1] + translations.n[1], 
       n[2] + translations.n[2]);
}

Vector3
Vector3::rotate(const Vector3 rotations)
// point *this is rotated one axis at a time, in order x, y, z
{
   double sinTheta, cosTheta;
   double x1, y1, z1, x2, y2, z2;

   // x rotation
   sinTheta = sin(rotations.n[0]);
   cosTheta = cos(rotations.n[0]);
   y1 = n[1]*cosTheta - n[2]*sinTheta;
   z1 = n[1]*sinTheta + n[2]*cosTheta;

   // y axis rotation
   sinTheta = sin(rotations.n[1]);
   cosTheta = cos(rotations.n[1]);
   z2 = z1*cosTheta - n[0]*sinTheta;
   x1 = z1*sinTheta + n[0]*cosTheta;

   // z rotation
   sinTheta = sin(rotations.n[2]);
   cosTheta = cos(rotations.n[2]);
   x2 = x1*cosTheta - y1*sinTheta;
   y2 = x1*sinTheta + y1*cosTheta;

   return Vector3(x2, y2, z2);
}

double 
Vector3::distanceTo(const Vector3 vector) const
{
   double dx = vector.n[0] - n[0];
   double dy = vector.n[1] - n[1];
   double dz = vector.n[2] - n[2];
   return sqrt(dx*dx + dy*dy + dz*dz);
}

Vector3 
Vector3::rotationsToXXX(const Vector3 target) const
// Returns the rotation vector which will transform *this to align with target.
// *this and target are both direction vectors.
{
   // first calculate x rotation required to reposition *this 
   // so that its endpoint lies on the line y == target.n[1] 
   // in the same sector as the target endpoint
   double lengthSquared = n[1]*n[1] + n[2]*n[2];
   double newX = n[0]; // stays the same
   double newY = target.n[1];
   double newZ = sqrt(lengthSquared - newY*newY);
   // place newZ in same sector as target endpoint
   if (target.n[2] < 0)
      newZ = -newZ;
   // calculate required x rotation 
   double xRotation = atan2(newZ, newY) - atan2(n[2], n[1]);

   // now calculate y rotation required to align
   // the new vector with the target
   double yRotation = atan2(target.n[0], target.n[2]) - atan2(newX, newZ);

   // thats it - no z rotation required for this case.

   return Vector3(xRotation, yRotation, 0);
}

Matrix4
Vector3::alignWith(const Vector3 target) const
// Returns a transformation which will align the orientation of *this with
// target. This is achievable by rotation about two axes only. After
// normalizing we try rotations about candidate pairs of axes, considering
// the plane projection geometry of the vectors.

{
   double thisLength2, targetLength2;  // projected length squared
   double rotationX, rotationY, rotationZ;
   Vector3 thisCopyA, thisCopyB, targetCopy;
   Vector3 xAxis, yAxis, zAxis;
   Matrix4 firstRotation, secondRotation;

   xAxis = Vector3(1.0, 0.0, 0.0);
   yAxis = Vector3(0.0, 1.0, 0.0);
   zAxis = Vector3(0.0, 0.0, 1.0);

   thisCopyA = *this;
   thisCopyA.normalize();
   targetCopy = target;
   targetCopy.normalize();

   // try first rotation about x axis, second about y axis -
   thisLength2 = 
      thisCopyA[VY]*thisCopyA[VY] + thisCopyA[VZ]*thisCopyA[VZ];
   targetLength2 = 
      targetCopy[VY]*targetCopy[VY] + targetCopy[VZ]*targetCopy[VZ];
   if (thisLength2 >= targetLength2) {  // x rotation ok
      thisCopyB[VX] = thisCopyA[VX]; // unchanged
      thisCopyB[VY] = targetCopy[VY];
      thisCopyB[VZ] = sqrt(thisLength2 - thisCopyB[VY]*thisCopyB[VY]);
      if (targetCopy[VZ] < 0)  // preserve sign
         thisCopyB[VZ] = -thisCopyB[VZ];
      rotationX = atan2(thisCopyB[VZ], thisCopyB[VY]) - 
                      atan2(thisCopyA[VZ], thisCopyA[VY]);
      rotationY = -(atan2(targetCopy[VZ], targetCopy[VX]) - 
                      atan2(thisCopyB[VZ], thisCopyB[VX]));
      firstRotation = rotation3D(xAxis, rotationX);
      secondRotation = rotation3D(yAxis, rotationY);
      return secondRotation*firstRotation;
   }

   // try first rotation about y axis, second about z axis -
   thisLength2 = 
      thisCopyA[VZ]*thisCopyA[VZ] + thisCopyA[VX]*thisCopyA[VX];
   targetLength2 = 
      targetCopy[VZ]*targetCopy[VZ] + targetCopy[VX]*targetCopy[VX];
   if (thisLength2 >= targetLength2) {  // x rotation ok
      thisCopyB[VY] = thisCopyA[VY]; // unchanged
      thisCopyB[VZ] = targetCopy[VZ];
      thisCopyB[VX] = sqrt(thisLength2 - thisCopyB[VZ]*thisCopyB[VZ]);
      if (targetCopy[VX] < 0)  // preserve sign
         thisCopyB[VX] = -thisCopyB[VX];
      rotationY = atan2(thisCopyB[VX], thisCopyB[VZ]) - 
                      atan2(thisCopyA[VX], thisCopyA[VZ]);
      rotationZ = -(atan2(targetCopy[VX], targetCopy[VY]) - 
                      atan2(thisCopyB[VX], thisCopyB[VY]));
      firstRotation = rotation3D(yAxis, rotationY);
      secondRotation = rotation3D(zAxis, rotationZ);
      return secondRotation*firstRotation;
   }

   // try first rotation about z axis, second about x axis -
   thisLength2 = 
      thisCopyA[VX]*thisCopyA[VX] + thisCopyA[VY]*thisCopyA[VY];
   targetLength2 = 
      targetCopy[VX]*targetCopy[VX] + targetCopy[VY]*targetCopy[VY];
   if (thisLength2 >= targetLength2) {  // x rotation ok
      thisCopyB[VZ] = thisCopyA[VZ]; // unchanged
      thisCopyB[VX] = targetCopy[VX];
      thisCopyB[VY] = sqrt(thisLength2 - thisCopyB[VX]*thisCopyB[VX]);
      if (targetCopy[VY] < 0)  // preserve sign
         thisCopyB[VY] = -thisCopyB[VY];
      rotationZ = atan2(thisCopyB[VY], thisCopyB[VX]) - 
                      atan2(thisCopyA[VY], thisCopyA[VX]);
      rotationX = -(atan2(targetCopy[VY], targetCopy[VZ]) - 
                      atan2(thisCopyB[VY], thisCopyB[VZ]));
      firstRotation = rotation3D(zAxis, rotationZ);
      secondRotation = rotation3D(xAxis, rotationX);
      return secondRotation*firstRotation;
   }

   return identity3D();
}

void 
Vector3::emitPovraySource
   (ofstream &povraySource, const Matrix4 &transformation) const
{
   Vector3 orig(n[0], n[1], n[2]);
   Vector3 trans = transformation * orig;
   povraySource << '<' << trans[0] << ", " << trans[1] << ", " << trans[2] << '>';
}


// FRIENDS

Vector3 operator - (const Vector3& a)
{  return Vector3(-a.n[VX],-a.n[VY],-a.n[VZ]); }

Vector3 operator + (const Vector3& a, const Vector3& b)
{ return Vector3(a.n[VX]+ b.n[VX], a.n[VY] + b.n[VY], a.n[VZ] + b.n[VZ]); }

Vector3 operator - (const Vector3& a, const Vector3& b)
{ return Vector3(a.n[VX]-b.n[VX], a.n[VY]-b.n[VY], a.n[VZ]-b.n[VZ]); }

Vector3 operator * (const Vector3& a, const double d)
{ return Vector3(d*a.n[VX], d*a.n[VY], d*a.n[VZ]); }

Vector3 operator * (const double d, const Vector3& a)
{ return a*d; }

Vector3 operator * (const Matrix4& a, const Vector3& v)
{ return a * Vector4(v); }

Vector3 operator * (const Vector3& v, Matrix4& a)
{ return a.transpose() * v; }

double operator * (const Vector3& a, const Vector3& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY] + a.n[VZ]*b.n[VZ]); }

Vector3 operator / (const Vector3& a, const double d)
{ double d_inv = 1./d; return Vector3(a.n[VX]*d_inv, a.n[VY]*d_inv,
  a.n[VZ]*d_inv); }

Vector3 operator ^ (const Vector3& a, const Vector3& b) {
    return Vector3
      (a.n[VY]*b.n[VZ] - a.n[VZ]*b.n[VY],
       a.n[VZ]*b.n[VX] - a.n[VX]*b.n[VZ],
       a.n[VX]*b.n[VY] - a.n[VY]*b.n[VX]);
}

int operator == (const Vector3& a, const Vector3& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]) && (a.n[VZ] == b.n[VZ]);
}

int operator != (const Vector3& a, const Vector3& b)
{ return !(a == b); }

ostream& operator << (ostream& s, Vector3& v)
{ return s << "| " << v.n[VX] << ' ' << v.n[VY] << ' ' << v.n[VZ] << " |"; }

istream& operator >> (istream& s, Vector3& v) {
    Vector3	v_tmp;
    char	c = ' ';

    while (isspace(c))
	s >> c;
    // The vectors can be formatted either as x y z or | x y z |
    if (c == '|') {
	s >> v_tmp[VX] >> v_tmp[VY] >> v_tmp[VZ];
	while (s >> c && isspace(c)) ;
//	if (c != '|')
//	    s.set(_bad);     JS NOTE: _bad causes error when fstream.h included
	}
    else {
	s.putback(c);
	s >> v_tmp[VX] >> v_tmp[VY] >> v_tmp[VZ];
	}
    if (s)
	v = v_tmp;
    return s;
}

void swap(Vector3& a, Vector3& b)
{ Vector3 tmp(a); a = b; b = tmp; }

Vector3 min(const Vector3& a, const Vector3& b)
{ return Vector3(MIN(a.n[VX], b.n[VX]), MIN(a.n[VY], b.n[VY]), MIN(a.n[VZ],
  b.n[VZ])); }

Vector3 max(const Vector3& a, const Vector3& b)
{ return Vector3(MAX(a.n[VX], b.n[VX]), MAX(a.n[VY], b.n[VY]), MAX(a.n[VZ],
  b.n[VZ])); }

Vector3 prod(const Vector3& a, const Vector3& b)
{ return Vector3(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY], a.n[VZ] * b.n[VZ]); }

