
// C++ Vector and Matrix Algebra Routines from Graphics Gems IV
// Author: Jean-Francois DOUE
// Version 3.1 --- October 1993
// Minor modifications by John Storrs

#include <math.h>
#include <iostream>
#include <fstream>
#include <ctype.h>

#include "Vector3.H"
#include "Vector4.H"
#include "Matrix4.H"

using namespace std;

// Vector4 Class Member Functions

// CONSTRUCTORS

Vector4::Vector4() {}

Vector4::Vector4(const double x, const double y, const double z, const double w)
{ n[VX] = x; n[VY] = y; n[VZ] = z; n[VW] = w; }

Vector4::Vector4(const double d)
{  n[VX] = n[VY] = n[VZ] = n[VW] = d; }

Vector4::Vector4(const Vector4& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = v.n[VW]; }

Vector4::Vector4(const Vector3& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = 1.0; }

Vector4::Vector4(const Vector3& v, const double d)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ];  n[VW] = d; }


// ASSIGNMENT OPERATORS

Vector4& Vector4::operator = (const Vector4& v)
{ n[VX] = v.n[VX]; n[VY] = v.n[VY]; n[VZ] = v.n[VZ]; n[VW] = v.n[VW];
return *this; }

Vector4& Vector4::operator += ( const Vector4& v )
{ n[VX] += v.n[VX]; n[VY] += v.n[VY]; n[VZ] += v.n[VZ]; n[VW] += v.n[VW];
return *this; }

Vector4& Vector4::operator -= ( const Vector4& v )
{ n[VX] -= v.n[VX]; n[VY] -= v.n[VY]; n[VZ] -= v.n[VZ]; n[VW] -= v.n[VW];
return *this; }

Vector4& Vector4::operator *= ( const double d )
{ n[VX] *= d; n[VY] *= d; n[VZ] *= d; n[VW] *= d; return *this; }

Vector4& Vector4::operator /= ( const double d )
{ double d_inv = 1./d; n[VX] *= d_inv; n[VY] *= d_inv; n[VZ] *= d_inv;
  n[VW] *= d_inv; return *this; }

double& Vector4::operator [] ( int i) {
    if (i < VX || i > VW)
	V_ERROR("Vector4 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

double Vector4::length()
{ return sqrt(length2()); }

double Vector4::length2()
{ return n[VX]*n[VX] + n[VY]*n[VY] + n[VZ]*n[VZ] + n[VW]*n[VW]; }

Vector4& Vector4::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }

Vector4& Vector4::apply(V_FCT_PTR fct)
{ n[VX] = (*fct)(n[VX]); n[VY] = (*fct)(n[VY]); n[VZ] = (*fct)(n[VZ]);
n[VW] = (*fct)(n[VW]); return *this; }


// FRIENDS

Vector4 operator - (const Vector4& a)
{ return Vector4(-a.n[VX],-a.n[VY],-a.n[VZ],-a.n[VW]); }

Vector4 operator + (const Vector4& a, const Vector4& b)
{ return Vector4(a.n[VX] + b.n[VX], a.n[VY] + b.n[VY], a.n[VZ] + b.n[VZ],
  a.n[VW] + b.n[VW]); }

Vector4 operator - (const Vector4& a, const Vector4& b)
{  return Vector4(a.n[VX] - b.n[VX], a.n[VY] - b.n[VY], a.n[VZ] - b.n[VZ],
   a.n[VW] - b.n[VW]); }

Vector4 operator * (const Vector4& a, const double d)
{ return Vector4(d*a.n[VX], d*a.n[VY], d*a.n[VZ], d*a.n[VW] ); }

Vector4 operator * (const double d, const Vector4& a)
{ return a*d; }

Vector4 operator * (const Matrix4& a, const Vector4& v) {
    #define ROWCOL(i) a.v[i].n[0]*v.n[VX] + a.v[i].n[1]*v.n[VY] \
    + a.v[i].n[2]*v.n[VZ] + a.v[i].n[3]*v.n[VW]
    return Vector4(ROWCOL(0), ROWCOL(1), ROWCOL(2), ROWCOL(3));
    #undef ROWCOL
}

Vector4 operator * (const Vector4& v, Matrix4& a)
{ return a.transpose() * v; }

double operator * (const Vector4& a, const Vector4& b)
{ return (a.n[VX]*b.n[VX] + a.n[VY]*b.n[VY] + a.n[VZ]*b.n[VZ] +
  a.n[VW]*b.n[VW]); }

Vector4 operator / (const Vector4& a, const double d)
{ double d_inv = 1./d; return Vector4(a.n[VX]*d_inv, a.n[VY]*d_inv, a.n[VZ]*d_inv,
  a.n[VW]*d_inv); }

int operator == (const Vector4& a, const Vector4& b)
{ return (a.n[VX] == b.n[VX]) && (a.n[VY] == b.n[VY]) && (a.n[VZ] == b.n[VZ])
  && (a.n[VW] == b.n[VW]); }

int operator != (const Vector4& a, const Vector4& b)
{ return !(a == b); }

ostream& operator << (ostream& s, Vector4& v)
{ return s << "| " << v.n[VX] << ' ' << v.n[VY] << ' ' << v.n[VZ] << ' '
  << v.n[VW] << " |"; }

istream& operator >> (istream& s, Vector4& v) {
    Vector4	v_tmp;
    char	c = ' ';

    while (isspace(c))
	s >> c;
    // The vectors can be formatted either as x y z w or | x y z w |
    if (c == '|') {
	s >> v_tmp[VX] >> v_tmp[VY] >> v_tmp[VZ] >> v_tmp[VW];
	while (s >> c && isspace(c)) ;
//	if (c != '|')
//	    s.set(_bad);     JS NOTE: _bad causes error when fstream.h included
	}
    else {
	s.putback(c);
	s >> v_tmp[VX] >> v_tmp[VY] >> v_tmp[VZ] >> v_tmp[VW];
	}
    if (s)
	v = v_tmp;
    return s;
}

void swap(Vector4& a, Vector4& b)
{ Vector4 tmp(a); a = b; b = tmp; }

Vector4 min(const Vector4& a, const Vector4& b)
{ return Vector4(MIN(a.n[VX], b.n[VX]), MIN(a.n[VY], b.n[VY]), MIN(a.n[VZ],
  b.n[VZ]), MIN(a.n[VW], b.n[VW])); }

Vector4 max(const Vector4& a, const Vector4& b)
{ return Vector4(MAX(a.n[VX], b.n[VX]), MAX(a.n[VY], b.n[VY]), MAX(a.n[VZ],
  b.n[VZ]), MAX(a.n[VW], b.n[VW])); }

Vector4 prod(const Vector4& a, const Vector4& b)
{ return Vector4(a.n[VX] * b.n[VX], a.n[VY] * b.n[VY], a.n[VZ] * b.n[VZ],
  a.n[VW] * b.n[VW]); }

