// name:     computebonecoords.cpp
// function: Computes bone reference coords in ring local coordinate systems.
// author:   John Storrs, Laboratory for Micro Enterprise.
// license:  Released under the terms of the GNU General Public License.
// date:     $Date: 2008-05-07 16:41:36 +0100 (Wed, 07 May 2008) $


#include <math.h>
#include <fstream>
#include <sstream>
#include <string>
#include "Vector3.H"
#include "Matrix4.H"
#include "fixatorframe.h"

using namespace std;

void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight);


int main(int argc, char **argv)
{
   int i, j, k, status;
   int ringHoleId[2][6];
   double ringNearest[2];
   double rotationX, rotationY, rotationZ;
   double translationX, translationY, translationZ;
   string fileName, fileSuffix, command;
   RingParameters ringParameters[2];
   Matrix4 transformation;
   StrutParameters strutParameters[6];
   FixatorFrame frame;
   XrayParameters xrayParameters;
   DeformityParameters deformityParameters;
   GraphicsParameters graphicsParameters;

   // get ring parameters from command line
   i = 1;
   for (j = 0; j < 2; j++) {
      ringParameters[j].innerRadius = atof(argv[i++]); 
      ringParameters[j].outerRadius =  atof(argv[i++]);
      ringParameters[j].thickness =  atof(argv[i++]);
      ringParameters[j].holeCircleRadius = atof(argv[i++]);
      ringParameters[j].holeCount = atoi(argv[i++]);
   }

   // get strut parameters from command line and set them
   for (j = 0; j < 6; j++) {
      strutParameters[j].minLength = atof(argv[i++]);
      strutParameters[j].maxLength = atof(argv[i++]);
      strutParameters[j].screwRadius = atof(argv[i++]);
      strutParameters[j].screwLength = atof(argv[i++]);
      strutParameters[j].nutRadius = atof(argv[i++]);
      strutParameters[j].nutLength = atof(argv[i++]);
      strutParameters[j].tubeRadius = atof(argv[i++]);
      strutParameters[j].tubeLength = atof(argv[i++]);
      strutParameters[j].jointRadius = atof(argv[i++]);
      strutParameters[j].jointOffset = atof(argv[i++]);
      frame.strut[j].setParameters(j, strutParameters[j]);
   }

   // now complete the ring parameters and set them
   for (j = 0; j < 2; j++) {
      for (k = 0; k < 6; k++) {
         ringParameters[j].jointRadius[k] = strutParameters[k].jointRadius;
         if (j == 0)
            ringParameters[j].jointOffset[k] = strutParameters[k].jointOffset;
         else
            ringParameters[j].jointOffset[k] = -strutParameters[k].jointOffset;
      }
      ringParameters[j].boneRadius = 2.0;
      frame.ring[j].setParameters(j, ringParameters[j]);
   }

   // get required strut length parameters from command line and set them
   for (j = 0; j < 4; j++)
      for (k = 0; k < 4; k++)
         transformation[j][k] = atof(argv[i++]);
   frame.ring[0].transformTo(transformation);
   for (j = 0; j < 4; j++)
      for (k = 0; k < 4; k++)
         transformation[j][k] = atof(argv[i++]);
   frame.ring[1].transformTo(transformation);

   // get xray parameters from command line and set them
   xrayParameters.xrayDistanceFrontal = atof(argv[i++]);
   xrayParameters.ring0NearestFrontal = atof(argv[i++]);
   xrayParameters.ring1NearestFrontal = atof(argv[i++]);
   xrayParameters.xrayDistanceLateral = atof(argv[i++]);
   xrayParameters.ring0NearestLateral = atof(argv[i++]);
   xrayParameters.ring1NearestLateral = atof(argv[i++]);
   frame.setXrayParameters(xrayParameters);

   // get required frame parameters from command line and set them
   for (j = 0; j < 6; j++) {
      ringHoleId[0][j] = atoi(argv[i++]) - 1;  // 1-based -> 0-based
      ringHoleId[1][j] = atoi(argv[i++]) - 1;
      frame.strut[j].connectRings(&frame.ring[0], &frame.ring[1], 
                                  ringHoleId[0][j], ringHoleId[1][j]);
   }

   // get deformity parameters from command line and set them
   deformityParameters.x1Ring0Frontal = atof(argv[i++]);
   deformityParameters.z1Ring0Frontal = atof(argv[i++]);
   deformityParameters.x2Ring0Frontal = atof(argv[i++]);
   deformityParameters.z2Ring0Frontal = atof(argv[i++]);
   deformityParameters.x1Ring1Frontal = atof(argv[i++]);
   deformityParameters.z1Ring1Frontal = atof(argv[i++]);
   deformityParameters.x2Ring1Frontal = atof(argv[i++]);
   deformityParameters.z2Ring1Frontal = atof(argv[i++]);
   deformityParameters.x1Ring0Lateral = atof(argv[i++]);
   deformityParameters.z1Ring0Lateral = atof(argv[i++]);
   deformityParameters.x2Ring0Lateral = atof(argv[i++]);
   deformityParameters.z2Ring0Lateral = atof(argv[i++]);
   deformityParameters.x1Ring1Lateral = atof(argv[i++]);
   deformityParameters.z1Ring1Lateral = atof(argv[i++]);
   deformityParameters.x2Ring1Lateral = atof(argv[i++]);
   deformityParameters.z2Ring1Lateral = atof(argv[i++]);
   deformityParameters.axialExtension = atof(argv[i++]);
   deformityParameters.axialRotation = atof(argv[i++]);
   frame.setDeformityParameters(deformityParameters);

   // get file name suffix from command line
   fileSuffix = argv[i++];

   // compute bone reference coordinates
   frame.computeBoneCoords();

   // set graphics parameters
   graphicsParameters.showOrigin = 0;
   graphicsParameters.showTarget = 1;
   graphicsParameters.showNearest = 1;
   graphicsParameters.showGrid = 1;
   graphicsParameters.showBones = 1;
   graphicsParameters.imageWidth = 280;  // mm
   graphicsParameters.imageHeight = 380;  // mm
   graphicsParameters.pixelsPerInch = 150;
   graphicsParameters.gridPitch = 20;  // mm
   frame.setGraphicsParameters(graphicsParameters);
   double widthPixels =
      graphicsParameters.imageWidth*graphicsParameters.pixelsPerInch/25.4;
   double heightPixels =
      graphicsParameters.imageHeight*graphicsParameters.pixelsPerInch/25.4;

   // transform frame to frontal view
   ringNearest[0] = xrayParameters.ring0NearestFrontal;
   ringNearest[1] = xrayParameters.ring1NearestFrontal;
   frame.transformView(ringNearest);

   // generate frontal view graphics
   ostringstream pathName1;
   pathName1 << "graphics/frontal" << fileSuffix << ".pov";
   ofstream povraySource1(pathName1.str().c_str());
   emitPovraySource(povraySource1, xrayParameters.xrayDistanceFrontal,
                    graphicsParameters.imageWidth, 
                    graphicsParameters.imageHeight);
   frame.emitPovraySource(povraySource1, xrayParameters.xrayDistanceFrontal);
   povraySource1.close();
   ostringstream smallFrontal;
   smallFrontal << "cd graphics; povray -Ifrontal" << fileSuffix << ".pov";
   smallFrontal << " +A +D +Q0 +W280 +H380";
   system(smallFrontal.str().c_str());
   ostringstream largeFrontal;
   largeFrontal << "cd graphics; povray -Ifrontal" << fileSuffix << ".pov";
   largeFrontal << " +A +D +Q0 +W" << widthPixels << " +H" << heightPixels
                << " +Ofrontal-large" << fileSuffix << ".png";
   system(largeFrontal.str().c_str());

   // transform frame to lateral view
   ringNearest[0] = xrayParameters.ring0NearestLateral;
   ringNearest[1] = xrayParameters.ring1NearestLateral;
   frame.transformView(ringNearest);

   // generate lateral view graphics
   ostringstream pathName2;
   pathName2 << "graphics/lateral" << fileSuffix << ".pov";
   ofstream povraySource2(pathName2.str().c_str());
   emitPovraySource(povraySource2, xrayParameters.xrayDistanceLateral,
                    graphicsParameters.imageWidth, 
                    graphicsParameters.imageHeight);
   frame.emitPovraySource(povraySource2, xrayParameters.xrayDistanceLateral);
   povraySource2.close();
   ostringstream smallLateral;
   smallLateral << "cd graphics; povray -Ilateral" << fileSuffix << ".pov";
   smallLateral << " +A +D +Q0 +W280 +H380";
   system(smallLateral.str().c_str());
   ostringstream largeLateral;
   largeLateral << "cd graphics; povray -Ilateral" << fileSuffix << ".pov";
   largeLateral << " +A +D +Q0 +W" << widthPixels << " +H" << heightPixels
                << " +Olateral-large" << fileSuffix << ".png";
   system(largeLateral.str().c_str());

   return 0;
}


void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight)
{
   povraySource << "global_settings { assumed_gamma 2.2 }" << endl;
   povraySource << "#include \"colors.inc\"" << endl;
   povraySource << "#include \"metals.inc\"" << endl;
   povraySource << "camera {" << endl;
   povraySource << "  perspective" << endl;
   povraySource << "  location <0.0, -" << xrayDistance << ", 0.0>" << endl;
   povraySource << "  direction <0.0, " << xrayDistance << ", 0.0>" << endl;
   povraySource << "  up <0, 0, " << imageHeight << ">" << endl;
   povraySource << "  right <" 
                << imageWidth << ", 0, 0>" << endl;  // right hand rule
   povraySource << "  sky <0, 0, 1>" << endl;
   povraySource << "  angle " 
                << 2*atan2(imageWidth/2, xrayDistance)*180/M_PI << endl;
   povraySource << "  look_at <0.0, 0.0, 0.0>" << endl;
   povraySource << "}" << endl;
   povraySource << "background { color red 1 green 1 blue 1 }" << endl;
   povraySource << "light_source { <-2000, -2000, 0> color Gray60}" << endl;
   povraySource << "#declare FixatorStrut =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.5 green 0.5 blue 0.5" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorGrid =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.3 green 0.3 blue 0.3" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorAxes =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorRing =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.8 green 0.8 blue 0.8" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorNearest =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 0.5 blue 1" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorTarget =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorOrigin=" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorBone =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorExtension =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorTrajectory =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 0.6 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
}
