// Copyright 2007 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// FixatorFrame - the assembled fixator frame

#include <math.h>
#include <values.h>
#include "primitive.h"
#include "artifact.h"
#include "fixatorframe.h"

Vector3 nearestPointToTwoLines
   (const Vector3 line1End1, const Vector3 line1End2,
    const Vector3 line2End1, const Vector3 line2End2);


FixatorFrame::FixatorFrame()
{
}


void
FixatorFrame::setXrayParameters(const XrayParameters &parameters)
{
   xrayParameters = parameters;
}


void
FixatorFrame::setDeformityParameters(const DeformityParameters &parameters)
{
   deformityParameters = parameters;
}


void
FixatorFrame::setGraphicsParameters(const GraphicsParameters &parameters)
{
   graphicsParameters = parameters;
}


// transformRingTo(): Absolute ring transformation.

void
FixatorFrame::transformRingTo(const int ringId, const Vector3 &rotation,
                              const Vector3 &translation)
{
   ring[ringId].transformTo(rotation, translation);
}


// transformRingBy(): Relative ring transformation.

void
FixatorFrame::transformRingBy(const int ringId, const Vector3 &rotation,
                              const Vector3 &translation)
{
   ring[ringId].transformBy(rotation, translation);
}


// transformBy(): Relative frame transformation.

void
FixatorFrame::transformBy(const Vector3 &rotation, const Vector3 &translation)
{
   Matrix4 transformation;

   ring[0].transformBy(rotation, translation);
   ring[1].transformBy(rotation, translation);

   transformation = Matrix4(identity3D(), rotation, translation);
   targetCoords = transformation*targetCoords;
}


void
FixatorFrame::transformBy(const Matrix4 &transformation)
{
   ring[0].transformBy(transformation);
   ring[1].transformBy(transformation);

   targetCoords = transformation*targetCoords;
}


// interpolate on real line segment containing min and max -
// index identifies interpolation position in range
double interpolate(int range, int index, double min, double max)
{
   return((index/double(range))*(max - min) + min);
}


// setStrutLengths() does a least squares binary search for a solution to the
// forward kinematics problem: given the strut lengths, find the frame
// geometry. Here ring[0] (ring 1 to the user) is fixed in space, and ring[1]
// (ring 2 to the user) is stepped through the 6-D search space. Note that we
// search using integer indices over a diminishing range, which are mapped to
// reals through calls of interpolate(). The algorithm finds a frame geometry
// which closely matches the given strut lengths, unless it fails to converge.
// When it terminates, ring[1] is left in the final position (the solution).

int
FixatorFrame::setStrutLengths(const double strutLength[6], 
                              const double approxZ)
{
   int index[6], target[6], range, iterations;
   double min[6], max[6], newMin, newMax;
   double precision, diff, squares, leastSquares;

   const int maxIterations = 100;
   const double tiny = 1.0e-1;

   // initial search space
   min[0] = min[1] = min[2] = -M_PI/4;  // x, y, z rotation
   max[0] = max[1] = max[2] = M_PI/4;
   min[3] = min[4] = -100.0;  // x, y translation
   max[3] = max[4] = 100.0;
   min[5] = approxZ - 50;
   max[5] = approxZ + 50;

   // search for least squares solution
   range = 5;
   iterations = 0;
   leastSquares = MAXDOUBLE;
   precision = 0.001;
   while (leastSquares > tiny) {
      iterations++;
      if (iterations > maxIterations) {
         cout << "not converging" << endl;
         return -1;
      }
      leastSquares = MAXDOUBLE;
      for (index[0] = 0; index[0] < range; index[0]++) {
         for (index[1] = 0; index[1] < range; index[1]++) {
            for (index[2] = 0; index[2] < range; index[2]++) {
               for (index[3] = 0; index[3] < range; index[3]++) {
                  for (index[4] = 0; index[4] < range; index[4]++) {
                     for (index[5] = 0; index[5] < range; index[5]++) {
                        transformRingTo
                          (1,
                           Vector3(
                              interpolate(range, index[0], min[0], max[0]), 
                              interpolate(range, index[1], min[1], max[1]), 
                              interpolate(range, index[2], min[2], max[2])),
                           Vector3(
                              interpolate(range, index[3], min[3], max[3]), 
                              interpolate(range, index[4], min[4], max[4]), 
                              interpolate(range, index[5], min[5], max[5])));
                        squares = 0.0;
                        for (int i = 0; i < 6; i++) {
                           diff = strutLength[i] - strut[i].length();
                           diff /= precision;
                           squares += diff*diff;
                        }
                        if (squares < leastSquares) {
                           leastSquares = squares;
                           for (int i = 0; i < 6; i++)
                              target[i] = index[i];
                        }
                     }
                  }
               }
            }
         }
      }
      for (int i = 0; i < 6; i++) {
         newMin = interpolate(range, target[i] - 1, min[i], max[i]);
         newMax = interpolate(range, target[i] + 1, min[i], max[i]);
         min[i] = newMin;
         max[i] = newMax;
      }

      cout << "leastSquares = " << leastSquares << endl;
      for (int i = 0; i < 3; i++) {
         cout << "val[i] = " 
              << interpolate(range, target[i], min[i], max[i])*180/M_PI << ", ";
      }
      for (int i = 3; i < 6; i++) {
         cout << "val[i] = " 
              << interpolate(range, target[i], min[i], max[i]) << ", ";
      }
      cout << "<br/>" << endl;
   }

   cout << "TRANSFORMATION = ";
   for (int i = 0; i < 3; i++) {
      cout << round(interpolate(
                 range, target[i], min[i], max[i])*(180.0/M_PI)/0.1)*0.1
           << " ";
   }
   for (int i = 3; i < 6; i++) {
      cout << round(interpolate(range, target[i], min[i], max[i])/0.1)*0.1
           << " ";
   }
   cout << "<br/>" << endl;

   return 0;
}

// transformView(): Transforms the frame so the view matches the x-ray image.
// ringNearest[0] and ringNearest[1] are values identifying the point on the
// perimeter of each ring nearest the x-ray source. The ring fixing holes
// provide a suitable real number scale for this, remembering that the first
// fixing hole number is 1, not 0. The x-ray target point is assumed to be
// midway between the two 'nearest' points. A new coordinate system is defined
// with the x-ray beam target line, assumed normal to the film cartridge, as
// the y axis. The x-ray source is at -ve y. The frame is transformed so that
// the x-ray target line passes through the frame centreline, the frame
// centreline is in the yz plane, and the rings are in contact with the film
// cartridge. The intersection of x-ray target line and frame centreline is the
// new coordinate system origin.  A deformity parameter measurement grid is
// marked on the y=0 plane.

void 
FixatorFrame::transformView(const double ringNearest[2])
{
   int i;
   double theta, xRotation, yRotation;
   Vector3 line, xAxis, yAxis, zAxis;
   Matrix4 transformation;

   xAxis = Vector3(1.0, 0.0, 0.0);
   yAxis = Vector3(0.0, 1.0, 0.0);
   zAxis = Vector3(0.0, 0.0, 1.0);

   // translate the frame so the origin is at the x-ray target point -
   // see FixatorRing::setParameters() etc to understand this
   for (i = 0; i < 2; i++) {
      theta = 2*M_PI*ringNearest[i]/ring[i].parameters.holeCount;
      ring[i].local.nearestCoords[0] = 
         ring[i].parameters.outerRadius*cos(theta);
      ring[i].local.nearestCoords[1] = 
         ring[i].parameters.outerRadius*sin(theta);
      ring[i].local.nearestCoords[2] = 0.0;
      ring[i].global.nearestCoords = 
         ring[i].transformation*ring[i].local.nearestCoords;
   }
   targetCoords = 
      (ring[0].global.nearestCoords + ring[1].global.nearestCoords)/2;
   transformBy(Vector3(0.0, 0.0, 0.0), -targetCoords);

   // rotate the frame so the line from the target point to the ring
   // centreline midpoint lies on the y axis, with target point at -ve y
   line = (ring[0].global.centreCoords + ring[1].global.centreCoords)/2 -
                                                                 targetCoords;
   transformation = line.alignWith(yAxis);
   transformBy(transformation);
   
   // translate the frame so the origin is at the ring centreline midpoint
   transformBy(Vector3(0.0, 0.0, 0.0), 
               -(ring[0].global.centreCoords + ring[1].global.centreCoords)/2);

   // rotate the frame about the y axis so ring centreline lies in the yz plane
   yRotation = 
      -atan2(ring[1].global.centreCoords[0], ring[1].global.centreCoords[2]); 
   transformBy(Vector3(0.0, yRotation, 0.0), Vector3(0.0, 0.0, 0.0));

   // translate the frame so the origin is back at the target point
   transformBy(Vector3(0.0, 0.0, 0.0), -targetCoords);

   // rotate the frame about the x axis so ring max y values are equal
   line = ring[1].yMaxCoords() - ring[0].yMaxCoords();
   xRotation = atan2(line[1], line[2]); 
   transformBy(Vector3(xRotation, 0.0, 0.0), Vector3(0.0, 0.0, 0.0));

   // translate the frame so the yMaxCoord() points are at y = 0.
   Vector3 yMaxCoords0 = ring[0].yMaxCoords();
   Vector3 yMaxCoords1 = ring[1].yMaxCoords();
   double yAdjustment = (yMaxCoords0[1] + yMaxCoords1[1])/2;
   transformBy(Vector3(0.0, 0.0, 0.0), 
               Vector3(0.0, -yAdjustment, 0.0));

   // frame transformation complete
}


// computeBoneCoords(): compute bone reference coordinates from deformity
// parameters.

void
FixatorFrame::computeBoneCoords()
{
   double ringNearest[2];
   Vector3 point1Ring1Frontal, point2Ring1Frontal,
           point1Ring2Frontal, point2Ring2Frontal,
           point1Ring1Lateral, point2Ring1Lateral,
           point1Ring2Lateral, point2Ring2Lateral;
   Vector3 xraySourceRing1Frontal, xraySourceRing2Frontal, 
           xraySourceRing1Lateral, xraySourceRing2Lateral;
   Matrix4 ring1Transform, ring2Transform;

   // transform frame to frontal view
   ringNearest[0] = xrayParameters.ring0NearestFrontal;
   ringNearest[1] = xrayParameters.ring1NearestFrontal;
   transformView(ringNearest);

   ring1Transform = ring[0].transformation.inverse();
   ring2Transform = ring[1].transformation.inverse();
   point1Ring1Frontal = 
      ring1Transform*Vector3(deformityParameters.x1Ring1Frontal, 0.0, 
                             deformityParameters.z1Ring1Frontal);
   point2Ring1Frontal = 
      ring1Transform*Vector3(deformityParameters.x2Ring1Frontal, 0.0, 
                             deformityParameters.z2Ring1Frontal);
   point1Ring2Frontal = 
      ring2Transform*Vector3(deformityParameters.x1Ring2Frontal, 0.0, 
                             deformityParameters.z1Ring2Frontal);
   point2Ring2Frontal = 
      ring2Transform*Vector3(deformityParameters.x2Ring2Frontal, 0.0, 
                             deformityParameters.z2Ring2Frontal);
   xraySourceRing1Frontal = 
      ring1Transform*Vector3(0.0, -xrayParameters.xrayDistanceFrontal, 0.0);
   xraySourceRing2Frontal = 
      ring2Transform*Vector3(0.0, -xrayParameters.xrayDistanceFrontal, 0.0);

   // transform frame to lateral view
   ringNearest[0] = xrayParameters.ring0NearestLateral;
   ringNearest[1] = xrayParameters.ring1NearestLateral;
   transformView(ringNearest);

   ring1Transform = ring[0].transformation.inverse();
   ring2Transform = ring[1].transformation.inverse();
   point1Ring1Lateral = 
      ring1Transform*Vector3(deformityParameters.x1Ring1Lateral, 0.0, 
                             deformityParameters.z1Ring1Lateral);
   point2Ring1Lateral = 
      ring1Transform*Vector3(deformityParameters.x2Ring1Lateral, 0.0, 
                             deformityParameters.z2Ring1Lateral);
   point1Ring2Lateral = 
      ring2Transform*Vector3(deformityParameters.x1Ring2Lateral, 0.0, 
                             deformityParameters.z1Ring2Lateral);
   point2Ring2Lateral = 
      ring2Transform*Vector3(deformityParameters.x2Ring2Lateral, 0.0, 
                             deformityParameters.z2Ring2Lateral);
   xraySourceRing1Lateral = 
      ring1Transform*Vector3(0.0, -xrayParameters.xrayDistanceLateral, 0.0);
   xraySourceRing2Lateral = 
      ring2Transform*Vector3(0.0, -xrayParameters.xrayDistanceLateral, 0.0);

   Vector3 point1Ring1, point2Ring1, point1Ring2, point2Ring2;

   point1Ring1 = 
      nearestPointToTwoLines(point1Ring1Frontal, xraySourceRing1Frontal,
                             point1Ring1Lateral, xraySourceRing1Lateral);
   point2Ring1 = 
      nearestPointToTwoLines(point2Ring1Frontal, xraySourceRing1Frontal,
                             point2Ring1Lateral, xraySourceRing1Lateral);
   point1Ring2 = 
      nearestPointToTwoLines(point1Ring2Frontal, xraySourceRing2Frontal,
                             point1Ring2Lateral, xraySourceRing2Lateral);
   point2Ring2 = 
      nearestPointToTwoLines(point2Ring2Frontal, xraySourceRing2Frontal,
                             point2Ring2Lateral, xraySourceRing2Lateral);

   ring[0].setLocalBoneCoords(point1Ring1, point2Ring1);
   ring[1].setLocalBoneCoords(point1Ring2, point2Ring2);
}


// correctDeformity(): This function provides progressive deformity correction.
// On each call the given fraction of the remaining deformity is corrected. So
// if the required total number of correction steps is 100, say, then 
// successive calls will use fraction arguments 1/100, 1/99, ... 1/2, 1.

void
FixatorFrame::correctDeformity(double fraction)
{
   Vector3 bone0 = ring[0].global.boneCoords[0] -
                       ring[0].global.boneCoords[1];
   Vector3 bone1 = ring[1].global.boneCoords[1] -
                       ring[1].global.boneCoords[0];

   // compute total rotation correction, normalising to +-PI
   Vector3 totalRotation = bone1.rotationsToXXX(bone0);
   for (int i = 0; i < 3; i++) {
      if (totalRotation[i] > 0.0) {
         while (totalRotation[i] > M_PI)
            totalRotation[i] -= 2*M_PI;
      } else {
         while (totalRotation[i] < -M_PI)
            totalRotation[i] += 2*M_PI;
      }
   }

   // apply fractional rotation and compute total translation correction
   Vector3 totalTranslation =
      ring[1].global.boneCoords[0] - ring[0].global.boneCoords[0];
   Vector3 targetCoords =
      ring[1].global.boneCoords[0] - totalTranslation*fraction;
   Vector3 rotation = totalRotation*fraction;
   transformRingBy(1, rotation, Vector3(0.0, 0.0, 0.0));
   Vector3 translation = targetCoords - ring[1].global.boneCoords[0];
   transformRingBy(1, Vector3(0.0, 0.0, 0.0), translation);
}


void
FixatorFrame::emitPovraySource(ofstream &povraySource, 
                               const double xrayDistance)
{
   const int xGridCount = (int)trunc(graphicsParameters.imageWidth/
                                     (2.0*graphicsParameters.gridPitch))*2;
   const int zGridCount = (int)trunc(graphicsParameters.imageHeight/
                                     (2.0*graphicsParameters.gridPitch))*2;
   const int xGridZMax = zGridCount*graphicsParameters.gridPitch/2;
   const int zGridXMax = xGridCount*graphicsParameters.gridPitch/2;
   int i;
   double xCoord, yCoord, zCoord, newYCoord, scale;
   Vector3 offsetCoords[2];
   Artifact frame("frame");
   Sphere target, origin, point[2];
   Cylinder centreline, xGrid[xGridCount], zGrid[zGridCount];
   Cylinder axes[2];

   for (i = 0; i < 6; i++)
      strut[i].emitPovraySource(povraySource);
   for (i = 0; i < 2; i++) {
      ring[i].showNearest = graphicsParameters.showNearest;
      ring[i].showBone = graphicsParameters.showBones;
      ring[i].emitPovraySource(povraySource);
   }

   if (graphicsParameters.showOrigin) {
      origin = Sphere(Vector3(0.0, 0.0, 0.0), 10, "FixatorOrigin");
      frame.add(&origin);
   }

   if (graphicsParameters.showTarget) {
      target = Sphere(targetCoords, 5, "FixatorTarget");
      frame.add(&target);
   }

   if (graphicsParameters.showBones) {
      // show bone corresponding points offset along ray 
      // in xray source direction so they are always visible
      if (xrayDistance > 0.0) {
         newYCoord = -300.0;
         for (i = 0; i < 2; i++) {
            xCoord = ring[i].global.boneCoords[0][0];
            yCoord = ring[i].global.boneCoords[0][1];
            zCoord = ring[i].global.boneCoords[0][2];
            scale = (xrayDistance + newYCoord)/(xrayDistance + yCoord);
            offsetCoords[i][0] = xCoord*scale;
            offsetCoords[i][1] = newYCoord;
            offsetCoords[i][2] = zCoord*scale;
            point[i] = Sphere(offsetCoords[i], 5*scale, "FixatorBone");
            frame.add(&point[i]);
         }
      }
   }

   if (graphicsParameters.showGrid) {
      for (i = 0; i < xGridCount; i++) {
         xGrid[i] = Cylinder
            (Vector3((i - xGridCount/2)*graphicsParameters.gridPitch, 
                     0.0, -xGridZMax), 
             Vector3((i - xGridCount/2)*graphicsParameters.gridPitch,
                     0.0, xGridZMax), 
             0.5, "FixatorGrid");
         frame.add(&xGrid[i]);
      }
      
      for (i = 0; i < zGridCount; i++) {
         zGrid[i] = Cylinder
            (Vector3(-zGridXMax, 0.0, 
                     (i - zGridCount/2)*graphicsParameters.gridPitch), 
             Vector3(zGridXMax, 0.0, 
                     (i - zGridCount/2)*graphicsParameters.gridPitch),
             0.5, "FixatorGrid");
         frame.add(&zGrid[i]);
      }

      axes[0] = Cylinder
         (Vector3(0.0, 0.0, -xGridZMax), 
          Vector3(0.0, 0.0, xGridZMax), 1.0, "FixatorAxes");
         frame.add(&axes[0]);
      axes[1] = Cylinder
         (Vector3(-zGridXMax, 0.0, 0.0), 
          Vector3(zGridXMax, 0.0, 0.0), 1.0, "FixatorAxes");
         frame.add(&axes[1]);
   }

   frame.emitPovraySource(povraySource, identity3D());
}


//===========================================================================

// SUPPORT CODE

/*
   CODE BY PAUL BOURKE, APRIL 1998
   See Internet article 'The Shortest Distance Between Two Lines in 3D'.

   Calculate the line segment PaPb that is the shortest route between
   two lines P1P2 and P3P4. Calculate also the values of mua and mub where
      Pa = P1 + mua (P2 - P1)
      Pb = P3 + mub (P4 - P3)
   Return FALSE if no solution exists.
*/

typedef struct {
   double x,y,z;
} XYZ;

int LineLineIntersect(
   XYZ p1,XYZ p2,XYZ p3,XYZ p4,XYZ *pa,XYZ *pb,
   double *mua, double *mub)
{
   XYZ p13,p43,p21;
   double d1343,d4321,d1321,d4343,d2121;
   double numer,denom;
   const double EPS = DBL_EPSILON;

   p13.x = p1.x - p3.x;
   p13.y = p1.y - p3.y;
   p13.z = p1.z - p3.z;
   p43.x = p4.x - p3.x;
   p43.y = p4.y - p3.y;
   p43.z = p4.z - p3.z;
   if (fabs(p43.x)  < EPS && fabs(p43.y)  < EPS && fabs(p43.z)  < EPS)
      return(-1);
   p21.x = p2.x - p1.x;
   p21.y = p2.y - p1.y;
   p21.z = p2.z - p1.z;
   if (fabs(p21.x)  < EPS && fabs(p21.y)  < EPS && fabs(p21.z)  < EPS)
      return(-1);

   d1343 = p13.x * p43.x + p13.y * p43.y + p13.z * p43.z;
   d4321 = p43.x * p21.x + p43.y * p21.y + p43.z * p21.z;
   d1321 = p13.x * p21.x + p13.y * p21.y + p13.z * p21.z;
   d4343 = p43.x * p43.x + p43.y * p43.y + p43.z * p43.z;
   d2121 = p21.x * p21.x + p21.y * p21.y + p21.z * p21.z;

   denom = d2121 * d4343 - d4321 * d4321;
   if (fabs(denom) < EPS)
      return(-1);
   numer = d1343 * d4321 - d1321 * d4343;

   *mua = numer / denom;
   *mub = (d1343 + d4321 * (*mua)) / d4343;

   pa->x = p1.x + *mua * p21.x;
   pa->y = p1.y + *mua * p21.y;
   pa->z = p1.z + *mua * p21.z;
   pb->x = p3.x + *mub * p43.x;
   pb->y = p3.y + *mub * p43.y;
   pb->z = p3.z + *mub * p43.z;

   return(0);
}


/* WRAPPER FOR PAUL BOURKE'S CODE */

Vector3 nearestPointToTwoLines
   (const Vector3 line1End1, const Vector3 line1End2,
    const Vector3 line2End1, const Vector3 line2End2)
{
   double mua, mub;
   XYZ p1, p2, p3, p4, a, b;

   p1.x = line1End1[0];
   p1.y = line1End1[1];
   p1.z = line1End1[2];
   p2.x = line1End2[0];
   p2.y = line1End2[1];
   p2.z = line1End2[2];
   p3.x = line2End1[0];
   p3.y = line2End1[1];
   p3.z = line2End1[2];
   p4.x = line2End2[0];
   p4.y = line2End2[1];
   p4.z = line2End2[2];

   LineLineIntersect(p1, p2, p3, p4, &a, &b, &mua, &mub);

   // TEMP
   double xAB = a.x - b.x;
   double yAB = a.y - b.y;
   double zAB = a.z - b.z;
   double lengthAB = sqrt(xAB*xAB + yAB*yAB + zAB*zAB);

   cout << "lengthAB = " << lengthAB << endl;

   return 
      Vector3((a.x + b.x)/2.0, (a.y + b.y)/2.0, (a.z + b.z)/2.0);
}
