// Copyright 2007 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// FixatorRing class

#include <math.h>
#include "primitive.h"
#include "artifact.h"
#include "fixatorring.h"
#include "fixatorstrut.h"


FixatorRing::FixatorRing()
{
   transformation = identity3D();
   local.centreCoords = Vector3(0.0, 0.0, 0.0);
   global.centreCoords = Vector3(0.0, 0.0, 0.0);
   showNearest = 0;
   showBone = 0;
}


void
FixatorRing::setParameters(const int id, const RingParameters &parameters)
{
   this->id = id;
   this->parameters = parameters;

   double theta;
   for (int holeId = 0; holeId < maxHoles; holeId++) {
      theta = 2*M_PI*holeId/parameters.holeCount;
      holeCoords[holeId][0] = parameters.holeCircleRadius*cos(theta);
      holeCoords[holeId][1] = parameters.holeCircleRadius*sin(theta);
      holeCoords[holeId][2] = 0.0;
   }
}


void
FixatorRing::connectStrut(FixatorStrut *strut, const int holeId)
{
   int strutId = strut->id;
   this->strut[strutId] = strut;

   local.ringCoords[strutId] = holeCoords[holeId];
   local.jointCoords[strutId] = holeCoords[holeId];
   local.jointCoords[strutId][2] += parameters.jointOffset[strutId];
   global.ringCoords[strutId] = transformation*local.ringCoords[strutId];
   global.jointCoords[strutId] = transformation*local.jointCoords[strutId];
}


void
FixatorRing::setLocalBoneCoords(const Vector3 &point0, const Vector3 &point1)
{
   local.boneCoords[0] = point0;
   local.boneCoords[1] = point1;
   global.boneCoords[0] = transformation*local.boneCoords[0];
   global.boneCoords[1] = transformation*local.boneCoords[1];
}


// TransformTo(): Absolute transformation. Note that the coordinate 
// transformation expression must be in the order 'transformation*coordinates'.
// This is not commutative.

void
FixatorRing::transformTo(const Vector3 &rotation, const Vector3 &translation)
{
   transformation = Matrix4(identity3D(), rotation, translation);

   // transform coordinates
   global.centreCoords = transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = transformation*local.ringCoords[i];
      global.jointCoords[i] = transformation*local.jointCoords[i];
   }
   global.nearestCoords = transformation*local.nearestCoords;
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = transformation*local.boneCoords[i];
   }
}


// TransformBy(): Relative transformation. Note that the coordinate 
// transformation expression must be in the order 'transformation*coordinates'.
// This is not commutative.

void
FixatorRing::transformBy(const Vector3 &rotation, const Vector3 &translation)
{
   transformation = Matrix4(transformation, rotation, translation);

   // transform coordinates
   global.centreCoords = transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = transformation*local.ringCoords[i];
      global.jointCoords[i] = transformation*local.jointCoords[i];
   }
   global.nearestCoords = transformation*local.nearestCoords;
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = transformation*local.boneCoords[i];
   }
}


void
FixatorRing::transformTo(const Matrix4 &transformation)
{
   this->transformation = transformation;

   // transform coordinates
   global.centreCoords = this->transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = this->transformation*local.ringCoords[i];
      global.jointCoords[i] = this->transformation*local.jointCoords[i];
   }
   global.nearestCoords = this->transformation*local.nearestCoords;
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = this->transformation*local.boneCoords[i];
   }
}


void
FixatorRing::transformBy(const Matrix4 &transformation)
{
   Matrix4 t = transformation;  // because of const
   this->transformation = t*this->transformation;

   // transform coordinates
   global.centreCoords = this->transformation*local.centreCoords;
   for (int i = 0; i < 6; i++) {
      global.ringCoords[i] = this->transformation*local.ringCoords[i];
      global.jointCoords[i] = this->transformation*local.jointCoords[i];
   }
   global.nearestCoords = this->transformation*local.nearestCoords;
   for (int i = 0; i < 2; i++) {
      global.boneCoords[i] = this->transformation*local.boneCoords[i];
   }
}


// yMaxCoords(): returns the point on the ring perimeter with the maximum y
// coordinate. Called in ../transformviews.cpp.

Vector3
FixatorRing::yMaxCoords()
{
   int i;
   double theta, yMax;
   Vector3 coords, maxCoords;

   theta = 0.0;
   yMax = -1000;
   for (i = 0; i < 10000; i++) {
      theta += 2*M_PI*i/10000.0;
      coords[0] = parameters.outerRadius*cos(theta);
      coords[1] = parameters.outerRadius*sin(theta);
      coords[2] = 0.0;
      coords = transformation*coords;
      if (coords[1] > yMax) {
         yMax = coords[1];
         maxCoords = coords;
      }
   }
   return maxCoords;
}


// emitPovraySource(): The model is constructed in local coordinate space,
// then transformed to global when emitPovraySource() for the ring is called.

void
FixatorRing::emitPovraySource(ofstream &povraySource)
{
   double jointClearance = 2.0;
   Vector3 supportCoords[6];
   Artifact ring("ring");
   OpenCylinder body;
   Cylinder support[6];
   Sphere joint[6];
   Sphere nearest;
   Cylinder bone;
   Sphere point;

   for (int i = 0; i < 6; i++) {
      supportCoords[i] = local.jointCoords[i];
      if (id == 0) 
         supportCoords[i][2] -= jointClearance;
      else
         supportCoords[i][2] += jointClearance;
   }
 
   if (id == 0) 
      body = OpenCylinder(Vector3(0.0, 0.0, 0.0), 
                          Vector3(0.0, 0.0, -parameters.thickness),
                          parameters.outerRadius, parameters.innerRadius,
                          "FixatorRing");
   else
      body = OpenCylinder(Vector3(0.0, 0.0, 0.0),
                          Vector3(0.0, 0.0, parameters.thickness),
                          parameters.outerRadius, parameters.innerRadius,
                          "FixatorRing");
   ring.add(&body);
   for (int i = 0; i < 6; i++) {
      support[i] = Cylinder(local.ringCoords[i], supportCoords[i],
                            parameters.jointRadius[i], "FixatorStrut");
      ring.add(&support[i]);
      joint[i] = Sphere(local.jointCoords[i], parameters.jointRadius[i],
                        "FixatorStrut");
      ring.add(&joint[i]);
   }
   if (showNearest) {
      nearest = Sphere(local.nearestCoords, 5, "FixatorNearest");
      ring.add(&nearest);
   }
   if (showBone) {
      bone = Cylinder(local.boneCoords[0], local.boneCoords[1],
                      parameters.boneRadius, "FixatorBone");
      ring.add(&bone);
      point = Sphere(local.boneCoords[0], 5, "FixatorBone");
      ring.add(&point);
   }

   ring.emitPovraySource(povraySource, transformation);
}

