// name:     setstrutlengths.cpp
// function: Computes hexapod fixator coordinates given strut lengths.
// author:   John Storrs, Laboratory for Micro Enterprise.
// license:  Released under the terms of the GNU General Public License.
// date:     $Date: 2008-06-22 22:29:58 +0100 (Sun, 22 Jun 2008) $


#include <math.h>
#include <fstream>
#include <sstream>
#include <string>
#include "Vector3.H"
#include "Matrix4.H"
#include "fixatorframe.h"

using namespace std;

void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight);


int main(int argc, char **argv)
{
   int i, j, k, status;
   int ringHoleId[2][6];
   double rotationX, rotationY, rotationZ;
   double translationX, translationY, translationZ;
   string fileName, fileSuffix, command;
   RingParameters ringParameters[2];
   Matrix4 transformation;
   StrutParameters strutParameters[6];
   FixatorFrame frame;
   GraphicsParameters graphicsParameters;

   // get ring parameters from command line
   i = 1;
   for (j = 0; j < 2; j++) {
      ringParameters[j].innerRadius = atof(argv[i++]); 
      ringParameters[j].outerRadius =  atof(argv[i++]);
      ringParameters[j].thickness =  atof(argv[i++]);
      ringParameters[j].holeCircleRadius = atof(argv[i++]);
      ringParameters[j].holeCount = atoi(argv[i++]);
   }

   // get strut parameters from command line and set them
   for (j = 0; j < 6; j++) {
      strutParameters[j].minLength = atof(argv[i++]);
      strutParameters[j].maxLength = atof(argv[i++]);
      strutParameters[j].screwRadius = atof(argv[i++]);
      strutParameters[j].screwLength = atof(argv[i++]);
      strutParameters[j].nutRadius = atof(argv[i++]);
      strutParameters[j].nutLength = atof(argv[i++]);
      strutParameters[j].tubeRadius = atof(argv[i++]);
      strutParameters[j].tubeLength = atof(argv[i++]);
      strutParameters[j].jointRadius = atof(argv[i++]);
      strutParameters[j].jointOffset = atof(argv[i++]);
      frame.strut[j].setParameters(j, strutParameters[j]);
   }

   // now complete the ring parameters and set them
   for (j = 0; j < 2; j++) {
      for (k = 0; k < 6; k++) {
         ringParameters[j].jointRadius[k] = strutParameters[k].jointRadius;
         if (j == 0) 
            ringParameters[j].jointOffset[k] = strutParameters[k].jointOffset;
         else
            ringParameters[j].jointOffset[k] = -strutParameters[k].jointOffset;
      }
      ringParameters[j].boneRadius = 2.0; 
      frame.ring[j].setParameters(j, ringParameters[j]);
   }

   // get required frame parameters from command line and set them
   for (j = 0; j < 6; j++) {
      ringHoleId[0][j] = atoi(argv[i++]) - 1;  // 1-based -> 0-based
      ringHoleId[1][j] = atoi(argv[i++]) - 1;
      frame.strut[j].connectRings(&frame.ring[0], &frame.ring[1], 
                                  ringHoleId[0][j], ringHoleId[1][j]);
   }

   // get strut length parameters from command line
   double strutLengths[6];
   double ringSpacing;
   for (j = 0; j < 6; j++) {
      strutLengths[j] = atof(argv[i++]);
   }
   ringSpacing = atof(argv[i++]);

   // get file name suffix from command line
   fileSuffix = argv[i++];

   status = frame.setStrutLengths(strutLengths, ringSpacing);
   if (status == -1)  // not converging
      return -1;

   // pass ring transformations to calling cgi script
   ostringstream pathName1;
   pathName1 << "transformations" << fileSuffix;
   ofstream transformations(pathName1.str().c_str());
   for (i = 0; i < 2; i++) {
      transformation = frame.ring[i].transformation;
      for (j = 0; j < 4; j++)
         for (k = 0; k < 4; k++)
            transformations << transformation[j][k] << " ";
      transformations << endl;
   }
   transformations.close();

   // set graphics parameters
   graphicsParameters.showOrigin = 0;
   graphicsParameters.showTarget = 0;
   graphicsParameters.showNearest = 0;
   graphicsParameters.showGrid = 0;
   graphicsParameters.showBones = 0;
   graphicsParameters.imageWidth = 280;  // mm
   graphicsParameters.imageHeight = 380;  // mm
   graphicsParameters.pixelsPerInch = 150;
   graphicsParameters.gridPitch = 20;  // mm
   frame.setGraphicsParameters(graphicsParameters);
   double widthPixels =
      graphicsParameters.imageWidth*graphicsParameters.pixelsPerInch/25.4;
   double heightPixels =
      graphicsParameters.imageHeight*graphicsParameters.pixelsPerInch/25.4;

   // generate front view graphic
   ostringstream pathName2;
   pathName2 << "graphics/frontview" << fileSuffix << ".pov";
   ofstream povraySource1(pathName2.str().c_str());
   emitPovraySource(povraySource1, 2000, 
                    graphicsParameters.imageWidth, 
                    graphicsParameters.imageHeight);
   frame.emitPovraySource(povraySource1, 0.0);
   povraySource1.close();
   ostringstream frontView;
   frontView << "cd graphics; povray -Ifrontview" << fileSuffix << ".pov";
   frontView << " +A +D +W280 +H380";
   system(frontView.str().c_str());

   // transform frame to side view
   Vector3 rotation = Vector3(0.0, 0.0, M_PI/2);
   Vector3 translation = Vector3(0.0, 0.0, 0.0);
   frame.transformRingBy(0, rotation, translation); 
   frame.transformRingBy(1, rotation, translation); 

   // generate side view graphic
   ostringstream pathName3;
   pathName3 << "graphics/sideview" << fileSuffix << ".pov";
   ofstream povraySource2(pathName3.str().c_str());
   emitPovraySource(povraySource2, 2000, 
                    graphicsParameters.imageWidth, 
                    graphicsParameters.imageHeight);
   frame.emitPovraySource(povraySource2, 0.0);
   povraySource2.close();
   ostringstream sideView;
   sideView << "cd graphics; povray -Isideview" << fileSuffix << ".pov";
   sideView << " +A +D +W280 +H380";
   system(sideView.str().c_str());

   return 0;
}


void emitPovraySource(ofstream &povraySource, const double xrayDistance,
                      const int imageWidth, const int imageHeight)
{
   povraySource << "global_settings { assumed_gamma 2.2 }" << endl;
   povraySource << "#include \"colors.inc\"" << endl;
   povraySource << "#include \"metals.inc\"" << endl;
   povraySource << "camera {" << endl;
   povraySource << "  perspective" << endl;
   povraySource << "  location <0.0, -" << xrayDistance << ", 0.0>" << endl;
   povraySource << "  direction <0.0, " << xrayDistance << ", 0.0>" << endl;
   povraySource << "  up <0, 0, " << imageHeight << ">" << endl;
   povraySource << "  right <" 
                << imageWidth << ", 0, 0>" << endl;  // right hand rule
   povraySource << "  sky <0, 0, 1>" << endl;
   povraySource << "  angle " 
                << 2*atan2(imageWidth/2, xrayDistance)*180/M_PI << endl;
   povraySource << "  look_at <0.0, 0.0, 0.0>" << endl;
   povraySource << "}" << endl;
   povraySource << "background { color red 1 green 1 blue 1 }" << endl;
   povraySource << "light_source { <-2000, -2000, 0> color Gray60}" << endl;
   povraySource << "#declare FixatorStrut =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.8 green 0.8 blue 0.8" << endl;
   povraySource << "  }" << endl;
   povraySource << "  finish {" << endl;
   povraySource << "    ambient 0.3" << endl;
   povraySource << "    brilliance 2" << endl;
   povraySource << "    diffuse 0.5" << endl;
   povraySource << "    metallic" << endl;
   povraySource << "    phong 1" << endl;
   povraySource << "    phong_size 300" << endl;
   povraySource << "    specular 0.8" << endl;
   povraySource << "    roughness 1/20" << endl;
   povraySource << "    reflection 0.2" << endl;
   povraySource << "    refraction 0" << endl;
   povraySource << "    ior 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorRing = FixatorStrut" << endl;
   povraySource << "#declare FixatorGrid =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0.3 green 0.3 blue 0.3" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorAxes =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorNearest =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 0.5 blue 1" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorTarget =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 1 green 0 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorOrigin=" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorBone =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorExtension =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 1 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
   povraySource << "#declare FixatorTrajectory =" << endl;
   povraySource << "texture {" << endl;
   povraySource << "  pigment {" << endl;
   povraySource << "    color red 0 green 0.6 blue 0" << endl;
   povraySource << "  }" << endl;
   povraySource << "}" << endl << endl;
}
